# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_m='config-false-prefixes'
_l='table-name-map'
_k='table-keys'
_j='app_ns'
_i='sztpd_meta'
_h='SELECT schema_name FROM information_schema.schemata;'
_g='key'
_f='cert'
_e='postgresql'
_d='Cannot delete: '
_c='[^/]*=[^/]*'
_b='[^:]*:'
_a='yang-library'
_Z='" does not exist.'
_Y=':tenants/tenant'
_X='opaque'
_W='global-root'
_V='db_ver'
_U='sqlite'
_T=':memory:'
_S='" already exists.'
_R='Node "'
_Q='Parent node ('
_P='.singletons'
_O='mysql'
_N='.*/'
_M=') does not exist.'
_L=':'
_K='jsob'
_J='row_id'
_I='ssl'
_H='pid'
_G='='
_F='=[^/]*'
_E='singletons'
_D=True
_C=False
_B='/'
_A=None
import os,re,sys,json,base64,pickle,yangson,binascii,pkg_resources
from .  import db_utils
import sqlalchemy as sa
from sqlalchemy.schema import CreateTable
from sqlalchemy.sql import and_
from dataclasses import dataclass
from enum import IntFlag
known_text=b'a secret message'
jsob_type=sa.JSON
class ContentType(IntFlag):CONFIG_TRUE=1;CONFIG_FALSE=2;CONFIG_ANY=3
@dataclass
class DatabasePath:data_path:str=_A;schema_path:str=_A;jsob_data_path:str=_A;table_name:str=_A;row_id:int=_A;inside_path:str=_A;path_segments:list=_A;jsob:dict=_A;node_ptr:dict=_A;prev_ptr:dict=_A
class NodeAlreadyExists(Exception):0
class NodeNotFound(Exception):0
class ParentNodeNotFound(Exception):0
class TooManyNodesFound(Exception):0
class InvalidResourceTarget(Exception):0
class CreateCallbackFailed(Exception):0
class ChangeCallbackFailed(Exception):0
class DeleteCallbackFailed(Exception):0
class DataAccessLayer:
	def __init__(A,db_url,cacert_param=_A,cert_param=_A,key_param=_A,yl_obj=_A,app_ns=_A,opaque=_A):
		F=yl_obj;E=key_param;D=cert_param;C=cacert_param;B=db_url;A.app_ns=_A;A.engine=_A;A.metadata=_A;A.leafrefs=_A;A.referers=_A;A.ref_stat_collectors=_A;A.global_root_id=_A;A.table_keys=_A;A.schema_path_to_real_table_name=_A;A.config_false_prefixes=_A;A.post_dal_callbacks=_A
		if F is _A:A._init(B,C,D,E)
		else:A.app_ns=app_ns;A._create(B,C,D,E,F,opaque)
		assert A.app_ns!=_A;assert A.engine!=_A;assert A.metadata!=_A;assert A.leafrefs!=_A;assert A.referers!=_A;assert A.ref_stat_collectors!=_A;assert A.global_root_id!=_A;assert A.table_keys!=_A;assert A.schema_path_to_real_table_name!=_A;assert A.config_false_prefixes!=_A
	def opaque(A):
		B=A.metadata.tables[A.schema_path_to_real_table_name[_E]]
		with A.engine.connect()as C:D=C.execute(sa.select([B.c.jsob]).where(B.c.name==_X));return D.first()[0]
	async def num_elements_in_list(A,data_path):
		B=data_path;C=re.sub(_F,'',B)
		if C!=B:assert NotImplementedError("Nested listed arren't supported yet.")
		D=A.schema_path_to_real_table_name[C];E=A.metadata.tables[D]
		with A.engine.connect()as F:G=F.execute(sa.select([sa.func.count()]).select_from(E));H=G.first()[0];return H
	async def get_tenant_name_for_admin(A,email_address):
		K='email-address';G=email_address
		if A.app_ns.endswith('x'):
			B=A.schema_path_to_real_table_name[_B+A.app_ns+':tenants/tenant/admin-accounts/admin-account'];C=A.metadata.tables[B]
			with A.engine.connect()as E:
				D=E.execute(sa.select([C.c.pid]).where(C.c[K]==G));F=D.first()
				if F!=_A:I=F[0];B=A.schema_path_to_real_table_name[_B+A.app_ns+_Y];H=A.metadata.tables[B];D=E.execute(sa.select([H.c.name]).where(H.c.row_id==I));J=D.first()[0];return J
		B=A.schema_path_to_real_table_name[_B+A.app_ns+':admin-accounts/admin-account'];C=A.metadata.tables[B]
		with A.engine.connect()as E:
			D=E.execute(sa.select([C.c.pid]).where(C.c[K]==G));F=D.first()
			if F!=_A:return _A
		raise NodeNotFound('Admin "'+G+_Z)
	async def get_tenant_name_for_global_key(A,table_name,k):
		B=table_name;C=A.schema_path_to_real_table_name[B];E=A.metadata.tables[C]
		with A.engine.connect()as F:
			D=F.execute(sa.select([E.c.pid]).where(getattr(E.c,A.table_keys[B])==k));G=D.first()
			if G==_A:raise NodeNotFound('key "'+k+'" in table "'+B+_Z)
			I=G[0];C=A.schema_path_to_real_table_name[_B+A.app_ns+_Y];H=A.metadata.tables[C];D=F.execute(sa.select([H.c.name]).where(H.c.row_id==I));J=D.first()[0];return J
	async def handle_get_opstate_request(B,data_path):
		A=data_path;assert A!='';assert not(A!=_B and A[-1]==_B)
		if A=='/ietf-yang-library:yang-library':
			E=B.schema_path_to_real_table_name[_E];C=B.metadata.tables[E]
			with B.engine.connect()as F:G=F.execute(sa.select([C.c.jsob]).where(C.c.name==_a));return G.first()[0]
		D=re.sub(_F,'',A)
		if 0:return await B._handle_get_data_request(A,D,ContentType.CONFIG_FALSE)
		else:return await B._handle_get_data_request(A,D,ContentType.CONFIG_ANY)
	async def handle_get_config_request(C,data_path):
		A=data_path;assert A!='';assert not(A!=_B and A[-1]==_B);B=re.sub(_F,'',A);F=await C._handle_get_data_request(A,B,ContentType.CONFIG_TRUE);D=re.sub(_N,'',B)
		if not D.startswith(C.app_ns+_L):D=C.app_ns+_L+D
		for E in C.config_false_prefixes:
			if E.startswith(B):
				if B==_B:G=E[1:]
				else:G=E.replace(B,D,1)
				I=_A;J=F
				def H(prev_ptr,curr_ptr,remainder_path):
					D=prev_ptr;B=remainder_path;A=curr_ptr;E=B.split(_B)
					for C in E:
						if type(A)==list:
							for F in A:H(A,F,B)
							return
						elif C in A:D=A;A=A[C];B=B.replace(C+_B,'',1)
						else:return
					D.pop(C)
				H(I,J,G)
			else:0
		return F
	async def _handle_get_data_request(C,data_path,schema_path,content_type):
		L='Node (';G=content_type;F=data_path;E=schema_path
		if G is ContentType.CONFIG_TRUE and any((E.startswith(A)for A in C.config_false_prefixes)):raise NodeNotFound(L+F+_M)
		with C.engine.connect()as H:
			A=C._get_dbpath_for_data_path(F,G,H)
			if A==_A:raise NodeNotFound(L+F+_M)
			B=re.sub(_N,'',E);I=B
			if B=='':D=A.node_ptr
			else:
				if not B.startswith(C.app_ns+_L):B=C.app_ns+_L+B
				D={}
				if A.table_name!=_E and I==A.inside_path:D[B]=[];D[B].append(A.node_ptr);A.node_ptr=D[B][0]
				else:D[B]=A.node_ptr
			J=C._get_list_of_direct_subtables_for_schema_path(E)
			for K in J:await C._recursively_attempt_to_get_data_from_subtable(K,A.row_id,E,A.node_ptr,G,H)
		return D
	async def _recursively_attempt_to_get_data_from_subtable(E,subtable_name,pid,jsob_schema_path,jsob_iter,content_type,conn):
		J=content_type;I=jsob_schema_path;C=subtable_name
		if not ContentType.CONFIG_FALSE in J:
			if any((C.startswith(A)for A in E.config_false_prefixes)):return
			else:0
		else:0
		Q=E._find_rows_in_table_having_pid(C,pid,conn);K=Q.fetchall()
		if len(K)==0:return
		B=jsob_iter
		if I==_B:M=C[1:]
		else:assert I.startswith(_B+E.app_ns);M=C.replace(I+_B,'')
		D=re.sub(_N,'',C);G=re.sub(D+'$','',M)
		if G!='':
			G=G[:-1];R=G.split(_B)
			for L in R:
				try:B=B[L]
				except KeyError as W:assert ContentType.CONFIG_FALSE in J;B[L]={};B=B[L]
		if any((C.startswith(A)for A in E.config_false_prefixes)):
			B[D]=[];S=E.schema_path_to_real_table_name[C];T=E.metadata.tables[S]
			for F in K:
				H={}
				for A in T.c:
					if A.name!=_J and A.name!=_H:
						if type(A.type)is sa.sql.sqltypes.DateTime:H[A.name]=F[A.name].strftime('%Y-%m-%dT%H:%M:%SZ')
						elif type(A.type)is sa.sql.sqltypes.JSON or type(A.type)is sa.sql.sqltypes.PickleType:
							if F[A.name]is not _A and not(type(F[A.name])is list and len(F[A.name])==0):H[A.name]=F[A.name]
						elif F[A.name]is not _A:H[A.name]=F[A.name]
				B[D].append(H)
		else:
			N=0
			for O in K:
				P=O[_K];D=next(iter(P))
				if D not in B:B[D]=[]
				B[D].append(P.pop(D));U=E._get_list_of_direct_subtables_for_schema_path(C)
				for V in U:await E._recursively_attempt_to_get_data_from_subtable(V,O[0],C,B[D][N],J,conn)
				N+=1
	async def handle_post_config_request(A,data_path,request_body,create_callbacks,change_callbacks,opaque):
		E=request_body;B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B);assert type(E)==dict
		with A.engine.begin()as C:
			F=A._get_dbpath_for_data_path(B,ContentType.CONFIG_TRUE,C)
			if F==_A:raise ParentNodeNotFound(_Q+B+_M)
			await A._handle_post_config_request(F,E,create_callbacks,change_callbacks,opaque,C)
			if A.post_dal_callbacks is not _A:
				for D in A.post_dal_callbacks:
					try:await D[0](D[1],C,D[2])
					except Exception as G:A.post_dal_callbacks=_A;raise G
				A.post_dal_callbacks=_A
	async def _handle_post_config_request(C,parent_dbpath,request_body,create_callbacks,change_callbacks,opaque,conn):
		P=change_callbacks;O=create_callbacks;M=opaque;H=request_body;D=conn;A=parent_dbpath;I=A.data_path;S=re.sub(_F,'',I);E=next(iter(H))
		if I==_B:B=E;F=_B+B
		else:B=re.sub(_b,'',E);F=I+_B+B
		T=re.sub(_F,'',F);G=C._get_table_name_for_schema_path(T);assert G!=_A
		if G==A.table_name:
			if B in A.node_ptr:raise NodeAlreadyExists(_R+B+_S)
			A.node_ptr[B]=H.pop(E);U=await C._recursively_post_subtable_data(A.row_id,F,A.node_ptr[B],A.jsob,A.jsob_data_path,O,M,D);C._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,D)
		else:
			if B not in A.node_ptr:A.node_ptr[B]=[];C._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,D)
			J={};assert len(H[E])==1;J[B]=H[E][0];Q=H[E][0][C.table_keys[G]];K={};K[_H]=A.row_id;K[C.table_keys[G]]=Q;K[_K]={};N=C.schema_path_to_real_table_name[G];L=C.metadata.tables[N]
			try:R=D.execute(L.insert().values(**K))
			except sa.exc.IntegrityError:raise NodeAlreadyExists(_R+B+_S)
			F+=_G+str(Q);U=await C._recursively_post_subtable_data(R.inserted_primary_key[0],F,J[B],J,F,O,M,D);N=C.schema_path_to_real_table_name[G];L=C.metadata.tables[N];D.execute(L.update().where(L.c.row_id==R.inserted_primary_key[0]).values(jsob=J))
		V=re.sub(_F,'',I)
		if V in P:
			for W in P[S]:await W(I,A.jsob,A.jsob_data_path,M)
	async def _recursively_post_subtable_data(B,pid,data_path,req_body_iter,jsob,jsob_data_path,create_callbacks,opaque,conn):
		N=jsob_data_path;M=jsob;J=opaque;F=data_path;E=conn;D=create_callbacks;C=req_body_iter;A=re.sub(_F,'',F)
		if type(C)==dict:
			if A in D:
				for O in D[A]:await O(F,M,N,J,E)
			for P in C.copy():K=F+_B+P;R=await B._recursively_post_subtable_data(pid,K,C[P],M,N,D,J,E)
		elif type(C)==list and A in B.table_keys:
			if A in B.table_keys:
				while C:
					G=C.pop(0);assert type(G)==dict;S=re.sub(_N,'',A);L={};L[S]=G;H=B.table_keys[A];I={};I[_H]=pid;I[H]=G[H];I[_K]=L;T=B.schema_path_to_real_table_name[A];U=B.metadata.tables[T]
					try:Q=E.execute(U.insert().values(**I))
					except sa.exc.IntegrityError as V:raise NodeAlreadyExists(_R+H+'" with value "'+I[H]+_S)
					K=F+_G+str(G[H]);R=await B._recursively_post_subtable_data(Q.inserted_primary_key[0],K,G,L,K,D,J,E);B._update_jsob_for_row_id_in_table(A,Q.inserted_primary_key[0],L,E)
				assert type(C)==list;assert len(C)==0
		elif A in D:
			for O in D[A]:await O(F,M,N,J,E)
		if A in B.table_keys:return _D
		return _C
	async def handle_post_opstate_request(A,data_path,request_body):
		P='Unrecognized resource schema path: ';I=request_body;B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B);C=re.sub(_F,'',B);F=next(iter(I))
		if C==_B:G=F;D=_B+G
		else:G=re.sub(_b,'',F);D=C+_B+G
		J=A._get_table_name_for_schema_path(D)
		if J!=D:raise NodeNotFound(P+D)
		E=I[F];K=re.findall(_c,B)
		if len(K)==0:E[_H]=A.global_root_id
		else:
			L=A._get_table_name_for_schema_path(C)
			if L==_A:raise ParentNodeNotFound(P+C)
			M=K[-1].split(_G)
			with A.engine.connect()as H:E[_H]=A._get_row_id_for_key_in_table(L,M[1],H)
			if E[_H]==_A:raise ParentNodeNotFound('Nonexistent parent resource: '+B)
		N=A.schema_path_to_real_table_name[J];O=A.metadata.tables[N]
		with A.engine.connect()as H:Q=H.execute(O.insert().values(**E))
	async def handle_put_config_request(A,data_path,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque):
		B=data_path;assert B!='';assert not(B!=_B and B[-1]==_B)
		with A.engine.begin()as D:
			await A._handle_put_config_request(B,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque,D)
			if A.post_dal_callbacks is not _A:
				for C in A.post_dal_callbacks:
					try:await C[0](C[1],D,C[2])
					except Exception as E:A.post_dal_callbacks=_A;raise E
				A.post_dal_callbacks=_A
	async def _handle_put_config_request(C,data_path,request_body,create_callbacks,change_callbacks,delete_callbacks,opaque,conn):
		L=delete_callbacks;I=opaque;H=change_callbacks;G=create_callbacks;E=data_path;D=conn;B=request_body;J=re.sub(_F,'',E)
		if E==_B:A=C._get_dbpath_for_data_path(_B,ContentType.CONFIG_ANY,D);assert A!=_A;await C.recursive_compare_and_put(A.row_id,_B,B,A.node_ptr,_A,A,G,H,L,I,D);N=C.schema_path_to_real_table_name[A.table_name];K=C.metadata.tables[N];D.execute(K.update().where(K.c.row_id==A.row_id).values(jsob=A.jsob));return
		if J.count(_B)>1:assert type(B)==dict;assert len(B)==1;F=next(iter(B));assert J.endswith(F.replace(C.app_ns+_L,'',1));O=B;B=B[F]
		if type(B)==list:assert len(B)==1;B=B[0]
		A=C._get_dbpath_for_data_path(E,ContentType.CONFIG_ANY,D)
		if A==_A:
			assert E!=_B;M,P=E.rsplit(_B,1)
			if M=='':M=_B
			A=C._get_dbpath_for_data_path(M,ContentType.CONFIG_ANY,D)
			if A==_A:raise ParentNodeNotFound(_Q+M+_M)
			if J.count(_B)>1:await C._handle_post_config_request(A,O,G,H,I,D)
			else:await C._handle_post_config_request(A,B,G,H,I,D)
			C._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,D);return
		if J.count(_B)==1 and J!=_B:
			if J in C.table_keys:F=next(iter(B));assert type(B[F])==list;assert len(B[F])==1;B=B[F][0];await C.recursive_compare_and_put(A.row_id,E,B,A.node_ptr,A.prev_ptr,A,G,H,L,I,D)
			else:F=next(iter(B));B=B[F];await C.recursive_compare_and_put(A.row_id,E,B,A.node_ptr,A.prev_ptr,A,G,H,L,I,D)
		else:await C.recursive_compare_and_put(A.row_id,E,B,A.node_ptr,A.prev_ptr,A,G,H,L,I,D)
		N=C.schema_path_to_real_table_name[A.table_name];K=C.metadata.tables[N];D.execute(K.update().where(K.c.row_id==A.row_id).values(jsob=A.jsob))
	async def recursive_compare_and_put(D,pid,data_path,req_body_iter,dbpath_curr_ptr,dbpath_prev_ptr,dbpath,create_callbacks,change_callbacks,delete_callbacks,opaque,conn):
		W=dbpath_prev_ptr;O=delete_callbacks;N=create_callbacks;M=pid;K=change_callbacks;J=opaque;I=conn;H=dbpath;F=dbpath_curr_ptr;C=req_body_iter;B=data_path;assert type(C)==type(F);E=re.sub(_F,'',B)
		if B==_B:0
		if type(C)==dict:
			P=set(list(C.keys()));Q=set(list(F.keys()))
			for A in [A for A in P if A not in Q]:
				if type(C[A])==list:U=_B+A if B==_B else B+_B+A;await D.recursive_compare_and_put(M,U,C[A],[],_A,H,N,K,O,J,I);F[A]=[]
				else:
					E=re.sub(_F,'',B);R=_B+A if E==_B else E+_B+A;assert type(C)!=list;G=_B+A if B==_B else B+_B+A;F[A]=C[A];a=await D._recursively_post_subtable_data(M,G,C[A],H.jsob,H.jsob_data_path,N,J,I)
					if a==1:assert type(C)==dict;assert type(C[A])==list;assert len(C[A])==0;C.pop(A);F.pop(A)
			for A in Q-P:
				if _C:U=_B+A if B==_B else B+_B+A;await D.recursive_compare_and_put(M,U,[],F[A],F,H,N,K,O,J,I);del F[A]
				else:
					R=_B+A if E==_B else E+_B+A
					if R in D.config_false_prefixes:0
					else:G=_B+A if B==_B else B+_B+A;await D._recursively_delete_subtable_data(H.row_id,G,F[A],O,J,I);del F[A]
			X=_C
			for A in Q&P:
				G=_B+A if B==_B else B+_B+A;R=re.sub(_F,'',G)
				if R in D.schema_path_to_real_table_name:0
				b=await D.recursive_compare_and_put(M,G,C[A],F[A],F,H,N,K,O,J,I)
				if b==_D:X=_D
				if R in D.schema_path_to_real_table_name:0
			if Q-P or(P-Q or X==_D):
				if E in K:
					for V in K[E]:await V(B,H.jsob,H.jsob_data_path,J)
			return _C
		elif type(C)==list and E in D.table_keys:
			Y=[A[D.table_keys[E]]for A in C];T=set(Y);S=set([A[0]for A in D._get_keys_in_table_having_pid(E,M,I)])
			for A in [A for A in Y if A not in S]:assert B!=_B;G=B+_G+A;c=[B for B in C if B[D.table_keys[E]]==A][0];d=[c];g=await D._recursively_post_subtable_data(M,B,d,_A,_A,N,J,I)
			for A in S-T:G=B+_G+A;h,Z=B.rsplit(_B,1);assert Z!='';await D._recursively_delete_subtable_data(H.row_id,G,W[Z],O,J,I)
			for A in S&T:G=B+_G+A;L=D._get_dbpath_for_data_path(G,ContentType.CONFIG_TRUE,I);assert L!=_A;e=[B for B in C if B[D.table_keys[E]]==A][0];await D.recursive_compare_and_put(L.row_id,G,e,L.node_ptr,L.prev_ptr,L,N,K,O,J,I);D._update_jsob_for_row_id_in_table(L.table_name,L.row_id,L.jsob,I)
			if T-S or S-T:return _D
			return _C
		else:
			if F!=C:
				f=re.sub('^.*/','',B);W[f]=C
				if E in K:
					for V in K[E]:await V(B,H.jsob,H.jsob_data_path,J)
			else:0
			return _C
		raise NotImplementedError('logic never reaches this point')
	async def handle_put_opstate_request(E,data_path,request_body):
		C=data_path;B=request_body;assert C!='';assert not(C!=_B and C[-1]==_B)
		with E.engine.begin()as G:
			D,F=C.rsplit(_B,1);assert F!=''
			if D=='':D=_B
			A=E._get_dbpath_for_data_path(D,ContentType.CONFIG_ANY,G)
			if A==_A:raise ParentNodeNotFound(_Q+D+_M)
			if type(B)==str:assert type(A.node_ptr)==dict;A.node_ptr[F]=B
			else:assert type(A.node_ptr)==dict and type(B)==dict;A.node_ptr[F]=B[next(iter(B))]
			E._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,G)
	async def handle_delete_config_request(A,data_path,delete_callbacks,change_callbacks,opaque):
		B=data_path;assert B!='';assert B!=_B;assert B[-1]!=_B
		with A.engine.begin()as D:
			await A._handle_delete_config_request(B,delete_callbacks,change_callbacks,opaque,D)
			if A.post_dal_callbacks is not _A:
				for C in A.post_dal_callbacks:
					try:await C[0](C[1],D,C[2])
					except Exception as E:A.post_dal_callbacks=_A;raise E
				A.post_dal_callbacks=_A
	async def _handle_delete_config_request(E,data_path,delete_callbacks,change_callbacks,opaque,conn):
		I=opaque;H=change_callbacks;F=conn;C=data_path;D,G=C.rsplit(_B,1)
		if D=='':D=_B
		A=E._get_dbpath_for_data_path(D,ContentType.CONFIG_TRUE,F)
		if A==_A:raise NodeNotFound(_d+C)
		if _G in G:B,O=G.rsplit(_G,1)
		else:B=G
		assert type(A.node_ptr)==dict
		if B not in A.node_ptr:raise NodeNotFound('Cannot delete '+C+'.')
		await E._recursively_delete_subtable_data(A.row_id,C,A.node_ptr[B],delete_callbacks,I,F)
		if type(A.node_ptr[B])==list:
			K=re.sub(_F,'',C);L=E._find_rowids_in_table_having_pid(K,A.row_id,F);M=L.fetchall()
			if len(M)==0:assert type(A.node_ptr[B])==list;assert len(A.node_ptr[B])==0;A.node_ptr.pop(B)
		else:A.node_ptr.pop(B)
		E._update_jsob_for_row_id_in_table(A.table_name,A.row_id,A.jsob,F);J=re.sub(_F,'',D)
		if J in H:
			for N in H[J]:await N(D,A.jsob,A.jsob_data_path,I)
	async def _recursively_delete_subtable_data(A,pid,data_path,curr_data_iter,delete_callbacks,opaque,conn):
		H=opaque;G=conn;F=delete_callbacks;E=pid;D=data_path;B=curr_data_iter;C=re.sub(_F,'',D)
		if type(B)==list and C in A.table_keys:
			assert B==[];R,M=D.rsplit(_B,1)
			async def J(pid,data_path,delete_callbacks,opaque,conn):
				D=conn;C=data_path;F=re.sub(_F,'',C);B=A._get_dbpath_for_data_path(C,ContentType.CONFIG_TRUE,D)
				if B==_A:raise NodeNotFound(_d+C)
				G=next(iter(B.jsob));H=B.jsob[G];await A._recursively_delete_subtable_data(B.row_id,C,H,delete_callbacks,opaque,D);I=A.schema_path_to_real_table_name[F];E=A.metadata.tables[I];J=D.execute(sa.delete(E).where(E.c.row_id==B.row_id));assert J.rowcount==1
			if _G in M:await J(E,D,F,H,G)
			else:
				N=[B[0]for B in A._get_keys_in_table_having_pid(C,E,G)]
				for O in N:I=D+_G+O;await J(E,I,F,H,G)
		elif type(B)==list and any((C.startswith(B)for B in A.config_false_prefixes)):assert B==[];P=A.schema_path_to_real_table_name[C];K=A.metadata.tables[P];S=G.execute(sa.delete(K).where(K.c.pid==E))
		elif type(B)==dict:
			for L in B.keys():assert D!=_B;I=D+_B+L;await A._recursively_delete_subtable_data(E,I,B[L],F,H,G)
		else:0
		if not(type(B)==list and C in A.table_keys)and not any((C.startswith(B)for B in A.config_false_prefixes)):
			if C in F:
				for Q in F[C]:await Q(D,H)
	def _find_rows_in_table_having_pid(B,table_name,pid,conn):C=B.schema_path_to_real_table_name[table_name];A=B.metadata.tables[C];D=conn.execute(sa.select([A]).where(A.c.pid==pid).order_by(A.c.row_id));return D
	def _find_rowids_in_table_having_pid(B,table_name,pid,conn):C=B.schema_path_to_real_table_name[table_name];A=B.metadata.tables[C];D=conn.execute(sa.select([A.c.row_id]).where(A.c.pid==pid).order_by(A.c.row_id));return D
	def _get_keys_in_table_having_pid(A,table_name,pid,conn):B=table_name;D=A.schema_path_to_real_table_name[B];C=A.metadata.tables[D];E=conn.execute(sa.select([getattr(C.c,A.table_keys[B])]).where(C.c.pid==pid));return E
	def _get_list_of_direct_subtables_for_schema_path(D,schema_path):
		A=schema_path
		if A!=_B:assert A[-1]!=_B;A+=_B
		C=[]
		for B in sorted(D.schema_path_to_real_table_name.keys()):
			if str(B).startswith(A):
				if not any((A for A in C if str(B).startswith(A+_B))):
					if str(B)!=_B:C.append(str(B))
					else:0
				else:0
		return C
	def _get_row_id_for_key_in_table(B,table_name,key,conn):
		C=table_name;E=B.schema_path_to_real_table_name[C];D=B.metadata.tables[E];F=conn.execute(sa.select([D.c.row_id]).where(getattr(D.c,B.table_keys[C])==key));A=F.fetchall();assert A is not _A
		if len(A)==0:return _A
		if len(A)>1:raise TooManyNodesFound()
		return A[0][0]
	def _get_jsob_iter_for_path_in_jsob(D,jsob,path):
		B=path;assert jsob!=_A;assert B[0]!=_B;A=jsob
		if B!='':
			for C in B.split(_B):
				if C!=''and C not in A:return _A
				A=A[C]
				if type(A)==list:assert len(A)==1;A=A[0]
		return A
	def _get_jsob_for_row_id_in_table(C,table_name,row_id,conn):
		F=row_id;D=table_name;K=C.schema_path_to_real_table_name[D];A=C.metadata.tables[K]
		if D in C.table_keys:E=conn.execute(sa.select([A.c.jsob]).where(A.c.row_id==F));G=E.first();assert G!=_A;return G[0]
		else:
			E=conn.execute(sa.select([A]).where(A.c.row_id==F));H=E.first();assert H!=_A;I=D.rsplit(_B,1)[1];J={I:{}}
			for B in A.c:
				if B.name!=_J and B.name!=_H:J[I][B.name]=H[B.name]
			return J
	def _insert_jsob_into_table(A,pid,table_name,new_jsob,conn):
		D=table_name;B=new_jsob;G=A.schema_path_to_real_table_name[D];H=A.metadata.tables[G];E=next(iter(B));C={};C[_H]=pid
		if D in A.table_keys:C[A.table_keys[D]]=B[E][A.table_keys[D]];C[_K]=B
		else:
			for F in B[E].keys():C[F]=B[E][F]
		I=conn.execute(H.insert().values(**C));return I.inserted_primary_key[0]
	def _update_jsob_for_row_id_in_table(A,table_name,row_id,new_jsob,conn):C=A.schema_path_to_real_table_name[table_name];B=A.metadata.tables[C];D=conn.execute(sa.update(B).where(B.c.row_id==row_id).values(jsob=new_jsob))
	def _get_table_name_for_schema_path(D,schema_path):
		B=len(_B);C=_E
		for A in D.schema_path_to_real_table_name.keys():
			if schema_path.startswith(A)and len(A)>B:B=len(A);C=A
		return C
	def _get_row_data_for_list_path(A,data_path,conn):
		B=data_path;assert B[0]==_B;assert B!=_B;assert B[-1]!=_B;G=B[1:].split(_B);assert _G in G[-1];D='';H=A.global_root_id
		for E in G:
			if _G in E:
				K,L=E.split(_G);D+=_B+K;I=A._get_table_name_for_schema_path(D);M=A.schema_path_to_real_table_name[I];C=A.metadata.tables[M];J=conn.execute(sa.select([C.c.row_id,C.c.pid]).where(and_(C.c.pid==H,getattr(C.c,A.table_keys[I])==L)));F=J.fetchone()
				if F==_A:return _A
				assert J.fetchone()==_A;H=F[_J]
			else:D+=_B+E
		return F
	def _get_dbpath_for_data_path(B,data_path,content_type,conn):
		E=conn;C=data_path;A=DatabasePath();A.data_path=C;A.schema_path=re.sub(_F,'',C)
		if _G not in C:A.jsob_data_path=_B
		else:A.jsob_data_path=re.sub('(.*=[^/]*).*','\\g<1>',C)
		A.table_name=B._get_table_name_for_schema_path(A.schema_path)
		if A.table_name==_A:return _A
		if ContentType.CONFIG_FALSE in content_type and any((A.table_name.startswith(C)for C in B.config_false_prefixes)):raise InvalidResourceTarget("RFC 8040 does not allow queries on lists directly and, because SZTPD doesn't support keys on 'config false' lists, it is never possible to query for 'dbpath.table_name' to be returned.  The 'val' layer should've rejected this query... ")
		D=re.findall(_c,C)
		if len(D)==0:A.row_id=B.global_root_id
		else:
			G=D[-1].split(_G)
			try:A.row_id=B._get_row_id_for_key_in_table(A.table_name,G[1],E)
			except TooManyNodesFound:
				H=B._get_row_data_for_list_path(A.jsob_data_path,E)
				if H is _A:A.row_id=_A
				else:A.row_id=H[_J]
			if A.row_id==_A:return _A
		if A.table_name.endswith(_E):assert A.table_name==_E;assert len(D)==0;A.inside_path=A.schema_path[1:]
		else:
			assert len(D)!=0;A.inside_path=G[0];I=re.sub('^'+A.table_name,'',A.schema_path)
			if I!='':A.inside_path+=I
		assert A.inside_path==''or A.inside_path[0]!=_B;A.jsob=B._get_jsob_for_row_id_in_table(A.table_name,A.row_id,E);A.node_ptr=A.jsob;A.prev_ptr=_A
		if A.inside_path=='':A.path_segments=[''];return A
		A.path_segments=A.inside_path.split(_B);J=''
		for F in A.path_segments:
			J+=_B+F
			if type(A.node_ptr)==list:A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[0]
			if F not in A.node_ptr:return _A
			else:A.prev_ptr=A.node_ptr;A.node_ptr=A.node_ptr[F]
		return A
	def _init(A,url,cacert_param,cert_param,key_param):
		J=key_param;I=cert_param;H=cacert_param;G=url
		if not(G.startswith('sqlite:///')or G.startswith(_O)or G.startswith(_e)):raise SyntaxError('The database url contains an unrecognized dialect.')
		E={}
		if H is not _A:
			E[_I]={};E[_I]['ca']=H
			if I is not _A:E[_I][_f]=I
			if J is not _A:E[_I][_g]=J
		A.engine=sa.create_engine(G,connect_args=E);A.db_schema=_A;A.table_keys={};A.config_false_prefixes={};A.schema_path_to_real_table_name={};A.leafrefs={};A.referers={};A.ref_stat_collectors={}
		if A.engine.url.database==_T or not db_utils.database_exists(A.engine.url,connect_args=E):A.engine=_A;raise NotImplementedError
		if A.engine.dialect.name==_U:A.schema_path_to_real_table_name[_B]=_E;A.schema_path_to_real_table_name[_E]=_E
		else:
			A.db_schema=A.engine.url.database;A.schema_path_to_real_table_name[_B]=A.db_schema.join(_P);A.schema_path_to_real_table_name[_E]=A.db_schema+_P;M=A.engine.execute(_h);K=M.fetchall();N=[K[A][0]for A in range(len(K))]
			if A.db_schema not in N:A.engine.execute(sa.schema.CreateSchema(A.db_schema));raise NotImplementedError
		A.metadata=sa.MetaData(bind=A.engine,schema=A.db_schema);A.metadata.reflect()
		for O in A.metadata.tables.values():
			for D in O.c:
				if type(D.type)is sa.sql.sqltypes.BLOB or type(D.type)is sa.sql.sqltypes.PickleType:D.type=sa.PickleType()
				if A.engine.dialect.name==_O and type(D.type)is sa.dialects.mysql.types.LONGTEXT:D.type=sa.JSON()
				elif type(D.type)is sa.sql.sqltypes.JSON:D.type=sa.JSON()
		B=A.metadata.tables[A.schema_path_to_real_table_name[_E]]
		with A.engine.connect()as F:
			C=F.execute(sa.select([B.c.jsob]).where(B.c.name==_i));L=C.first()[0]
			if L[_V]!=1:raise AssertionError('The database version ('+L[_V]+') is unexpected.')
			C=F.execute(sa.select([B.c.jsob]).where(B.c.name==_j));A.app_ns=C.first()[0];C=F.execute(sa.select([B.c.row_id]).where(B.c.name==_W));A.global_root_id=C.first()[0];C=F.execute(sa.select([B.c.jsob]).where(B.c.name==_k));A.table_keys=C.first()[0];C=F.execute(sa.select([B.c.jsob]).where(B.c.name==_l));A.schema_path_to_real_table_name=C.first()[0];C=F.execute(sa.select([B.c.jsob]).where(B.c.name==_m));A.config_false_prefixes=C.first()[0]
	def _create(A,url,cacert_param,cert_param,key_param,yl_obj,opaque):
		V='default startup endpoint';U='endpoint';T='SZTPD_TEST_PATH';S='_sztp_ref_stats_stmt';R='_sztp_globally_unique_stmt';J=key_param;I=cert_param;H=cacert_param;G='name';F=yl_obj;D={}
		if H is not _A:
			D[_I]={};D[_I]['ca']=H
			if I is not _A:D[_I][_f]=I
			if J is not _A:D[_I][_g]=J
		A.engine=sa.create_engine(url,connect_args=D)
		if A.engine.url.database!=_T and db_utils.database_exists(A.engine.url,connect_args=D):raise AssertionError('Database already exists (call init() first).')
		if A.engine.url.database!=_T:
			if A.engine.dialect.name==_O:db_utils.create_database(A.engine.url,encoding='utf8mb4',connect_args=D)
			else:db_utils.create_database(A.engine.url,connect_args=D)
		A.db_schema=_A
		if A.engine.dialect.name in(_O,_e):
			A.db_schema=str(A.engine.url.database);E=A.engine.execute(_h);K=E.fetchall();M=[K[A][0]for A in range(len(K))]
			if A.db_schema not in M:A.engine.execute('CREATE SCHEMA IF NOT EXISTS %s;'%A.db_schema)
		A.metadata=sa.MetaData(bind=A.engine,schema=A.db_schema);C=sa.Table(_E,A.metadata,sa.Column(_J,sa.Integer,primary_key=_D),sa.Column(_H,sa.Integer,unique=_D),sa.Column(G,sa.String(250),unique=_D),sa.Column(_K,jsob_type));A.metadata.create_all()
		with A.engine.connect()as B:B.execute(C.insert(),name=_i,jsob={_V:1});B.execute(C.insert(),name=_j,jsob=A.app_ns);B.execute(C.insert(),name=_a,jsob=F);B.execute(C.insert(),name=_X,jsob=opaque)
		A.table_keys={_E:G,_B:G};A.config_false_prefixes={};A.schema_path_to_real_table_name={};A.leafrefs={};A.referers={};A.ref_stat_collectors={}
		if A.engine.dialect.name==_U:A.schema_path_to_real_table_name[_B]=_E;A.schema_path_to_real_table_name[_E]=_E
		else:A.schema_path_to_real_table_name[_B]=A.db_schema+_P;A.schema_path_to_real_table_name[_E]=A.db_schema+_P
		def N(self,stmt,sctx):self.globally_unique=stmt.argument
		setattr(yangson.schemanode.SchemaNode,R,N);yangson.schemanode.SchemaNode._stmt_callback['wn-app:globally-unique']=R
		def O(self,stmt,sctx):self.ref_stats=stmt.argument
		setattr(yangson.schemanode.SchemaNode,S,O);yangson.schemanode.SchemaNode._stmt_callback['wn-app:ref-stats']=S;L=pkg_resources.resource_filename('sztpd','yang/')
		if os.environ.get(T):A.dm=yangson.DataModel(json.dumps(F),[os.environ.get(T),L])
		else:A.dm=yangson.DataModel(json.dumps(F),[L])
		A._gen_tables(A.dm.schema,_E)
		with A.engine.connect()as B:E=B.execute(C.insert(),name=_l,jsob=A.schema_path_to_real_table_name);E=B.execute(C.insert(),name=_k,jsob=A.table_keys);E=B.execute(C.insert(),name=_m,jsob=A.config_false_prefixes)
		if os.environ.get('SZTPD_TEST_DAL'):
			with A.engine.connect()as B:E=B.execute(C.insert().values(name=_W,jsob={}))
			A.global_root_id=E.inserted_primary_key[0]
		else:
			with A.engine.connect()as B:E=B.execute(C.insert().values(name=_W,jsob={A.app_ns+':transport':{'listen':{U:[]}},A.app_ns+':audit-log':{}}))
			A.global_root_id=E.inserted_primary_key[0];P=A.schema_path_to_real_table_name[_B+A.app_ns+':transport/listen/endpoint'];Q=A.metadata.tables[P]
			with A.engine.connect()as B:W=B.execute(Q.insert().values(pid=A.global_root_id,name=V,jsob={U:{G:V,'use-for':['native-interface'],'http':{'tcp-server-parameters':{'local-address':'127.0.0.1'}}}}))
	def _gen_tables(D,node,parent_table_name):
		R='ref_stats';G=parent_table_name;C=node
		if issubclass(type(C),yangson.schemanode.ListNode):
			B=[];B.append(sa.Column(_J,sa.Integer,primary_key=_D));N=D.schema_path_to_real_table_name[G];B.append(sa.Column(_H,sa.Integer,sa.ForeignKey(N+'.row_id'),index=_D,nullable=_C))
			if C.config==_D:
				if len(C.keys)>1:raise NotImplementedError('Only supports lists with at most one key.')
				E=C.get_child(*C.keys[0]);D.table_keys[C.data_path()]=E.name
				if type(E.type)==yangson.datatype.StringType:B.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.Uint32Type:B.append(sa.Column(E.name,sa.Integer,nullable=_C))
				elif type(E.type)==yangson.datatype.IdentityrefType:B.append(sa.Column(E.name,sa.String(250),nullable=_C))
				elif type(E.type)==yangson.datatype.UnionType:B.append(sa.Column(E.name,sa.String(250),nullable=_C))
				else:raise Exception('Unsupported key node type: '+str(type(E.type)))
				if hasattr(E,'globally_unique'):B.append(sa.UniqueConstraint(E.name))
				else:B.append(sa.UniqueConstraint(E.name,_H))
				B.append(sa.Column(_K,jsob_type,nullable=_C))
			if C.config==_C:
				assert hasattr(C,R)==_C
				for A in C.children:
					if issubclass(type(A),yangson.schemanode.LeafNode):
						if type(A.type)==yangson.datatype.StringType:
							if str(A.type)=='date-and-time(string)':B.append(sa.Column(A.name,sa.DateTime,index=_D,nullable=A.mandatory==_C or A.when!=_A))
							else:B.append(sa.Column(A.name,sa.String(250),index=_D,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.Uint16Type:B.append(sa.Column(A.name,sa.SmallInteger,index=_D,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.InstanceIdentifierType:B.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.LeafrefType:B.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.IdentityrefType:B.append(sa.Column(A.name,sa.String(250),nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.EnumerationType:B.append(sa.Column(A.name,sa.String(250),index=_D,nullable=A.mandatory==_C or A.when!=_A))
						elif type(A.type)==yangson.datatype.UnionType:
							J=0;S=_D
							for K in A.type.types:
								if issubclass(type(K),yangson.datatype.StringType):J+=1
								else:raise Exception('Unhandled union type: '+str(type(K)))
							if J==len(A.type.types):B.append(sa.Column(A.name,sa.String(250),index=_D,nullable=A.mandatory==_C or A.when!=_A))
							else:raise Exception('FIXME: not all union subtypes are stringafiable')
						else:raise Exception('Unhandled leaf type: '+str(type(A.type)))
					elif issubclass(type(A),yangson.schemanode.ChoiceNode):
						H=_D
						for I in A.children:
							assert type(I)==yangson.schemanode.CaseNode
							if len(I.children)>1:H=_C;break
							else:
								for O in I.children:
									if type(O)!=yangson.schemanode.LeafNode:H=_C;break
						if H==_D:B.append(sa.Column(A.name,sa.String(250),index=_D,nullable=A.mandatory==_C or A.when!=_A))
						else:B.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.AnydataNode):B.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.LeafListNode):B.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.ListNode):B.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.ContainerNode):B.append(sa.Column(A.name,jsob_type,nullable=A.mandatory==_C or A.when!=_A))
					elif issubclass(type(A),yangson.schemanode.NotificationNode):0
					else:raise Exception('Unhandled list child type: '+str(type(A)))
			P=re.sub('^/.*:','',C.data_path()).split(_B)
			if D.engine.dialect.name==_U:F=''
			else:F=D.db_schema+'.'
			for Q in P:F+=Q[0]
			while F in D.schema_path_to_real_table_name.values():F+='2'
			D.schema_path_to_real_table_name[C.data_path()]=F
			if D.db_schema is _A:L=sa.Table(F,D.metadata,*B)
			else:L=sa.Table(re.sub('^'+D.db_schema+'.','',F),D.metadata,*B)
			L.create();G=C.data_path()
		if C.config==_C and issubclass(type(C),yangson.schemanode.DataNode):
			M=C.data_path()
			if not any((M.startswith(A)for A in D.config_false_prefixes)):D.config_false_prefixes[M]=_D
		if hasattr(C,R):D.ref_stat_collectors[C.data_path()]=_A
		if issubclass(type(C),yangson.schemanode.InternalNode):
			if not(type(C)==yangson.schemanode.ListNode and C.config==_C)and not type(C)==yangson.schemanode.RpcActionNode and not type(C)==yangson.schemanode.NotificationNode:
				for A in C.children:D._gen_tables(A,G)


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_M='Unrecognized member: '
_L='The top-level node-identifier must be prefixed by a namespace followed by a colon.'
_K='Parent data node ('
_J='Input document must contain at least one top-level node.'
_I='Data node ('
_H='Unrecognized schema path: '
_G='Invalid data path: '
_F=":admin-accounts'"
_E='[/] missing-data: expected'
_D='Validation failed: '
_C=') does not exist.'
_B=True
_A='/'
import yangson,asyncio
class NodeAlreadyExists(Exception):0
class NodeNotFound(Exception):0
class ParentNodeNotFound(Exception):0
class InvalidDataPath(Exception):0
class InvalidInputDocument(Exception):0
class UnrecognizedInputNode(Exception):0
class UnrecognizedQueryParameter(Exception):0
class MissingQueryParameter(Exception):0
class NonexistentSchemaNode(Exception):0
class ValidationFailed(Exception):0
class ValidationLayer:
	def __init__(A,dm,dal):
		A.dm=dm;A.dal=dal;C=asyncio.get_event_loop();D=A.dal.handle_get_config_request(_A);E=C.run_until_complete(D);A.inst=A.dm.from_raw(E)
		try:A.inst.validate()
		except yangson.exceptions.SchemaError as B:assert str(B).startswith(_E);assert str(B).endswith(_F)
	async def reload(A):
		C=await A.dal.handle_get_config_request(_A);A.inst=A.dm.from_raw(C)
		try:A.inst.validate()
		except yangson.exceptions.SchemaError as B:assert str(B).startswith(_E);assert str(B).endswith(_F)
	async def handle_get_config_request(C,data_path,query_dict):
		A=data_path;assert A!='';assert not(A!=_A and A[-1]==_A)
		try:D=C.dm.parse_resource_id(A)
		except yangson.exceptions.UnexpectedInput as B:raise InvalidDataPath(_G+str(B))
		except yangson.exceptions.NonexistentSchemaNode as B:raise NonexistentSchemaNode(_H+A)
		try:E=C.inst.goto(D)
		except yangson.exceptions.NonexistentInstance as B:raise NodeNotFound(_I+A+_C)
	async def handle_post_config_request(I,data_path,query_dict,request_body):
		Y='=';Q=None;E=request_body;B=data_path;assert B!='';assert not(B!=_A and B[-1]==_A)
		if len(E)<1:raise InvalidInputDocument(_J)
		if len(E)>1:raise InvalidInputDocument('Input document must not have more than one top-level node.')
		try:R=I.dm.parse_resource_id(B)
		except yangson.exceptions.NonexistentSchemaNode as C:raise NonexistentSchemaNode('Unrecognized schema path for parent node: '+B)
		try:F=I.inst.goto(R)
		except yangson.exceptions.NonexistentInstance as C:raise ParentNodeNotFound(_K+B+_C)
		A=next(iter(E))
		if':'not in A:raise InvalidInputDocument(_L)
		S,G=A.split(':');K=F.schema_node;H=K.get_child(G,S)
		if H is Q:raise UnrecognizedInputNode('Input document contains unrecognized top-level node.')
		if not K.ns is Q:assert K.ns==H.ns
		if isinstance(H,yangson.schemanode.SequenceNode):
			if type(H)==yangson.schemanode.ListNode:
				T=H.keys[0];U=T[0]
				if type(E[A])!=list:raise InvalidInputDocument("Input node '"+G+"' not a 'list' node.")
				if len(E[A])!=1:raise InvalidInputDocument("Input 'list' node '"+G+"' must contain one element.")
				N=E[A][0];O=N[U]
			elif type(H)==yangson.schemanode.LeafListNode:raise NotImplementedError('Inserting into LeafListNode not implemented yet.')
			else:raise AssertionError('Logic cannot reach this point')
			if B==_A:D=A;L=_A+D+Y+O
			else:D=G;L=B+_A+D+Y+O
			try:V=I.dm.parse_resource_id(L)
			except yangson.exceptions.NonexistentSchemaNode as C:raise NonexistentSchemaNode('Unrecognized schema path for insertion node: '+L)
			try:I.inst.goto(V)
			except yangson.exceptions.NonexistentInstance as C:pass
			else:raise NodeAlreadyExists('Child data node ('+L+') already exists.')
			try:J=F[D]
			except yangson.exceptions.NonexistentInstance:J=F.put_member(D,yangson.instvalue.ArrayValue([]))
			assert isinstance(J.schema_node,yangson.schemanode.SequenceNode)
			if len(J.value)==0:
				try:W=J.update([N],raw=_B)
				except yangson.exceptions.RawMemberError as C:raise UnrecognizedInputNode('Incompatable node data. '+str(C))
				M=W.up()
			else:X=J[-1];M=X.insert_after(N,raw=_B).up()
		else:
			if B==_A:D=A
			else:D=G
			if D in F:raise NodeAlreadyExists('Node "'+D+'" already exists.')
			try:
				if K.ns==Q:M=F.put_member(A,E[A],raw=_B).up()
				else:M=F.put_member(G,E[A],raw=_B).up()
			except yangson.exceptions.RawMemberError as C:raise UnrecognizedInputNode(_M+str(C))
		P=M.top()
		try:P.validate()
		except Exception as C:raise ValidationFailed(_D+str(C))
		I.inst2=P
	async def handle_put_config_request(C,data_path,query_dict,request_body):
		D=request_body;B=data_path;assert B!='';assert not(B!=_A and B[-1]==_A)
		if len(D)<1:raise InvalidInputDocument(_J)
		G=next(iter(D))
		if':'not in G:raise InvalidInputDocument(_L)
		try:H=C.dm.parse_resource_id(B)
		except yangson.exceptions.UnexpectedInput as A:raise InvalidDataPath(_G+str(A))
		except yangson.exceptions.NonexistentSchemaNode as A:raise NonexistentSchemaNode(_H+B)
		try:E=C.inst.goto(H)
		except yangson.exceptions.NonexistentInstance as A:
			F=B.rsplit(_A,1)[0]
			if F=='':F=_A
			H=C.dm.parse_resource_id(F)
			try:E=C.inst.goto(H)
			except yangson.exceptions.NonexistentInstance as A:raise ParentNodeNotFound(_K+F+') does not exist. '+str(A))
			await C.handle_post_config_request(F,query_dict,D);return
		try:
			if B==_A:I=E.update(D,raw=_B)
			elif isinstance(E.schema_node,yangson.schemanode.SequenceNode):I=E.update(D[G][0],raw=_B)
			else:I=E.update(D[G],raw=_B)
		except yangson.exceptions.RawMemberError as A:raise UnrecognizedInputNode(_M+str(A))
		except Exception as A:raise NotImplementedError(str(type(A))+' = '+str(A))
		J=I.top()
		try:J.validate()
		except Exception as A:raise ValidationFailed(_D+str(A))
		C.inst2=J
	async def handle_delete_config_request(C,data_path):
		A=data_path;assert A!=''
		if A==_A:E=RootNode(ObjectValue({}),C.inst.schema_node,datetime.now())
		else:
			assert A[0]==_A;assert A[-1]!=_A
			try:J=C.dm.parse_resource_id(A)
			except yangson.exceptions.NonexistentSchemaNode as F:raise NonexistentSchemaNode('Unrecognized schema path for data node: '+A)
			try:D=C.inst.goto(J)
			except yangson.exceptions.NonexistentInstance as F:raise NodeNotFound(_I+A+_C)
			H=D.up()
			if type(D)==yangson.instance.ArrayEntry:
				B=H.delete_item(D.index)
				if len(B.value)==0:
					G=B.up()
					if isinstance(G.schema_node,yangson.schemanode.SequenceNode):I=G.delete_item(B.index);raise NotImplementedError('tested? list inside a list...')
					else:I=G.delete_item(B.name)
					B=I
			else:B=H.delete_item(D.name)
			E=B.top()
		try:E.validate()
		except Exception as F:raise ValidationFailed(_D+str(F))
		C.inst2=E


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

import sys,argparse
from .  import __version__,sztpd
def main(argv=None):A=argparse.ArgumentParser(prog='sztpd',formatter_class=argparse.RawDescriptionHelpFormatter,description='SZTPD implements the "bootstrap server" defined in RFC 8572.',epilog='\nExit status code: 0 on success, non-0 on error.  Error output goes to stderr.\n\nThe "cacert" argument is a filepath to a PEM file that contains one or more X.509\nCA certificates used to authenticate the RDBMS\'s TLS certificate.\n\nThe "key" and "cert" arguments are each a filepath to a PEM file that contains\nthe key and certificate that SZTPD should use to authenticate itself to the\nRDBMS.  These parameters must be specified together, and must be specified\nin conjunction with the "cacert" parameter.\n\nThe "database-url" argument has the form "<dialect>:<dialect-specific-path>".\nThree dialects are supported: "sqlite", "postgresql", and "mysql+pymysql".\nThe dialect-specific-path for each of these is described below.\n\nFor the "sqlite" dialect, <dialect-specific-path> follows the format\n"///<sqlite-path>", where <sqlite-path> can be one of:\n\n  :memory:    - an in-memory database (only useful for testing)\n  <filepath>  - an OS-specific filepath to a persisted database file\n\n  Examples:\n\n    $ sztpd sqlite:///:memory:                      (memory)\n    $ sztpd sqlite:///relative/path/to/sztpd.db     (unix)\n    $ sztpd sqlite:////absolute/path/to/sztpd.db    (unix)\n    $ sztpd sqlite:///C:\\path\\to\\sztpd.db           (windows)\n\nFor both the "postgresql" and "mysql+pymysql" dialects, <dialect-specific-path>\nfollows the format "//<user>[:<passwd>]@<host>:<port>/<database-name>".\n\n  Examples:\n\n    The following two examples assume the database is called "sztpd" and\n    that the database server listens on the loopback address with no TLS.\n\n      $ sztpd mysql+pymysql://user:pass@localhost:3306/sztpd\n      $ sztpd postgresql://user:pass@localhost:5432/sztpd\n\n\nPlease see the documentation for more information.\n');A.add_argument('-v','--version',help='show version number and exit.',action='version',version=__version__.__version__);A.add_argument('-C','--cacert',help='path to certificates used to authenticate the database (see below for details).');A.add_argument('-c','--cert',help='path to cert used to authenticate SZTPD to the database (see below for details).');A.add_argument('-k','--key',help='path to key used to authenticate SZTPD to the database (see below for details).');A.add_argument('database_url',help='see below for details.',metavar='database-url');B=A.parse_args();return sztpd.run(B.database_url,B.cacert,B.cert,B.key)
if __name__=='__main__':sys.exit(main())


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_E='content'
_D='certificates'
_C='certificate'
_B='contentType'
_A=None
import re,pem,base64,textwrap
from pyasn1.type import tag,univ
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5280
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
def gen_rc_errors(error_type,error_tag,error_app_tag=_A,error_path=_A,error_message=_A,error_info=_A):
	E=error_info;D=error_message;C=error_path;B=error_app_tag;A={};A['error-type']=error_type;A['error-tag']=error_tag
	if B is not _A:A['error-app-tag']=B
	if C is not _A:A['error-path']=C
	if D is not _A:A['error-message']=D
	if E is not _A:A['error-info']=E
	return{'ietf-restconf:errors':{'error':[A]}}
def multipart_pem_to_der_dict(multipart_pem):
	A={};E=pem.parse(bytes(multipart_pem,'utf-8'))
	for F in E:
		C=F.as_text().splitlines();D=base64.b64decode(''.join(C[1:-1]));B=re.sub('-----BEGIN (.*)-----','\\g<1>',C[0])
		if B not in A:A[B]=[D]
		else:A[B].append(D)
	return A
def der_dict_to_multipart_pem(der_dict):
	H='-----\n';C=der_dict;A='';D=C.keys()
	for B in D:
		E=C[B]
		for F in E:G=base64.b64encode(F).decode('ASCII');A+='-----BEGIN '+B+H;A+=textwrap.fill(G,64)+'\n';A+='-----END '+B+H
	return A
def ders_to_degenerate_cms_obj(cert_ders):
	B=rfc5652.CertificateSet().subtype(implicitTag=tag.Tag(tag.tagClassContext,tag.tagFormatSimple,0))
	for E in cert_ders:F,G=der_decoder.decode(E,asn1Spec=rfc5280.Certificate());assert not G;D=rfc5652.CertificateChoices();D[_C]=F;B[len(B)]=D
	A=rfc5652.SignedData();A['version']=1;A['digestAlgorithms']=rfc5652.DigestAlgorithmIdentifiers().clear();A['encapContentInfo']['eContentType']=rfc5652.id_data;A[_D]=B;C=rfc5652.ContentInfo();C[_B]=rfc5652.id_signedData;C[_E]=der_encoder.encode(A);return C
def degenerate_cms_obj_to_ders(cms_obj):
	A=cms_obj
	if A[_B]!=rfc5652.id_signedData:raise KeyError('unexpected content type: '+str(A[_B]))
	D,H=der_decoder.decode(A[_E],asn1Spec=rfc5652.SignedData());E=D[_D];B=[]
	for F in E:C=F[_C];assert type(C)==rfc5280.Certificate;G=der_encoder.encode(C);B.append(G)
	return B


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_G='tested?'
_F='$0$'
_E='wn-sztpd-0:device'
_D='device'
_C='activation-code'
_B='/'
_A=None
import gc,tracemalloc,os,re,json,signal,asyncio,datetime,functools,pkg_resources
from .  import yl
from .dal import DataAccessLayer,CreateCallbackFailed
from .rcsvr import RestconfServer
from passlib.hash import sha256_crypt
from .tenant import TenantViewHandler
from .rfc8572 import RFC8572ViewHandler
from .native import NativeViewHandler,Period,TimeUnit
loop=_A
sig=_A
def signal_handler(name):global loop;global sig;sig=name;loop.stop()
def run(db_url,cacert_param=_A,cert_param=_A,key_param=_A):
	d=':transport';c='SIGHUP';b=True;V='use-for';U='x';T='1';N=db_url;J=key_param;I=cert_param;H=cacert_param;global loop;global sig;A=_A;B=_A;Q=False
	if H is not _A and N.startswith('sqlite:'):print('The "sqlite" dialect does not support the "cacert" parameter.');return 1
	if(I or J)and not H:print('The "cacert" parameter must be specified whenever the "key" and "cert" parameters are specified.');return 1
	if(I is _A)!=(J is _A):print('The "key" and "cert" parameters must be specified together.');return 1
	try:A=DataAccessLayer(N,H,I,J)
	except (SyntaxError,AssertionError)as K:print(str(K));return 1
	except NotImplementedError as K:Q=b
	else:B=A.opaque()
	if Q==b:
		L=os.environ.get('SZTPD_MODE')
		if L!=_A:
			assert type(L)==str
			if L not in[T,U]:print("Unknown SZTPD_MODE value.  Must be '1' or 'x'.");return 1
			B=L
		else:
			print('');W=pkg_resources.resource_filename('sztpd','LICENSE.txt');R=open(W,'r');print(R.read());R.close();print('First time initialization.  Please accept the license terms.');print('');print('By entering "Yes" below, you agree to be bound to the terms and conditions contained on this screen with Watsen Networks.');print('');X=input('Please enter "Yes" or "No": ')
			if X!='Yes':print('');print('Thank you for your consideration.');print('');return 1
			print('');print('Modes:');print('  1 - single-tenant');print('  x - multi-tenant');print('');B=input('Please select mode: ')
			if B not in[T,U]:print('Unknown mode selection.  Please try again.');return 1
			print('');print("Running SZTPD in mode '"+B+"'. (No more output expected)");print('')
		try:A=DataAccessLayer(N,H,I,J,json.loads(getattr(yl,'nbi_'+B)()),'wn-sztpd-'+B,B)
		except Exception as K:raise K;return 1
	assert B!=_A;assert A!=_A;tracemalloc.start(25);loop=asyncio.get_event_loop();loop.add_signal_handler(signal.SIGHUP,functools.partial(signal_handler,name=c));loop.add_signal_handler(signal.SIGTERM,functools.partial(signal_handler,name='SIGTERM'));loop.add_signal_handler(signal.SIGINT,functools.partial(signal_handler,name='SIGINT'));loop.add_signal_handler(signal.SIGQUIT,functools.partial(signal_handler,name='SIGQUIT'))
	while sig is _A:
		M=[];E=A.handle_get_config_request(_B+A.app_ns+d);O=loop.run_until_complete(E)
		for D in O[A.app_ns+d]['listen']['endpoint']:
			if D[V][0]=='native-interface':
				C=NativeViewHandler(A,B,loop)
				if B=='0':F=_B+A.app_ns+':device'
				elif B==T:F=_B+A.app_ns+':devices/device'
				elif B==U:F=_B+A.app_ns+':tenants/tenant/devices/device'
				C.register_create_callback(F,_handle_device_created);Y=F+'/activation-code';C.register_change_callback(Y,_handle_device_act_code_changed);C.register_subtree_change_callback(F,_handle_device_subtree_changed);C.register_somehow_change_callback(F,_handle_device_somehow_changed);C.register_delete_callback(F,_handle_device_deleted);C.register_periodic_callback(Period(24,TimeUnit.Hours),datetime.datetime(2000,1,1,0),_check_expirations);P=RestconfServer(loop,A,D,C)
			elif D[V][0]=='tenant-interface':Z=TenantViewHandler(C);P=RestconfServer(loop,A,D,Z)
			else:assert D[V][0]=='rfc8572-interface';S=json.loads(getattr(yl,'sbi_rfc8572')());a=RFC8572ViewHandler(A,B,S,C);P=RestconfServer(loop,A,D,a,S)
			M.append(P);del D;D=_A
		del O;O=_A;loop.run_forever()
		for G in M:E=G.app.shutdown();loop.run_until_complete(E);E=G.runner.cleanup();loop.run_until_complete(E);E=G.app.cleanup();loop.run_until_complete(E);del G;G=_A
		del M;M=_A
		if sig==c:sig=_A
	loop.close();del A;return 0
async def _handle_device_created_post_sweep(watched_node_path,conn,opaque):
	k=':dynamic-callout';j='webhooks';i='verification-result';h='failure';g='=';f='tenant';e='function';d='functions';c='plugin';b='callback';a='ownership-authorization';W='verification-results';V='dynamic-callout';U='device-type';T='row_id';S='=[^/]*';J='wn-sztpd-rpcs:output';H=watched_node_path;B=conn;A=opaque;C=A.dal._get_row_data_for_list_path(H,B);D=re.sub(S,'',H);X=A.dal._get_jsob_for_row_id_in_table(D,C[T],B);K=_B+A.dal.app_ns+':device-types/device-type='+X[_D][U];C=A.dal._get_row_data_for_list_path(K,B);D=re.sub(S,'',K);L=A.dal._get_jsob_for_row_id_in_table(D,C[T],B)
	if a in L[U]:
		M=_B+A.dal.app_ns+':dynamic-callouts/dynamic-callout='+L[U][a][V]['reference'];C=A.dal._get_row_data_for_list_path(M,B);D=re.sub(S,'',M);E=A.dal._get_jsob_for_row_id_in_table(D,C[T],B)
		if b in E[V]:
			F=E[V][b];assert F[c]in A.plugins;N=A.plugins[F[c]];assert d in N;O=N[d];assert F[e]in O;Y=O[F[e]];I=H.split(_B)
			if I[2]==f:P=I[1].split(g)[1]
			else:P='not-applicable'
			Q=I[-1].split(g)[1];Z={'wn-sztpd-rpcs:input':{f:P,'serial-number':[Q]}};G=Y(Z);R=h
			if J in G:
				if W in G[J]:
					if i in G[J][W]:R=G[J][W][i][0]['result']
			if R==h:raise CreateCallbackFailed('Unable to verify ownership for device: '+Q)
		else:assert j in E[A.dal.app_ns+k][0];l=E[A.dal.app_ns+k][0][j];raise NotImplementedError('webhooks for ownership verification not implemented yet')
async def _handle_device_created(watched_node_path,jsob,jsob_data_path,nvh,conn):
	C=nvh;B=jsob;assert type(B)==dict
	if jsob_data_path==_B:assert _E in B;A=B[_E]
	else:assert _D in B;A=B[_D]
	if C.dal.post_dal_callbacks is _A:C.dal.post_dal_callbacks=[]
	C.dal.post_dal_callbacks.append((_handle_device_created_post_sweep,watched_node_path,C));A['lifecycle-statistics']={'nbi-access-stats':{'created':datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ'),'num-times-modified':0},'sbi-access-stats':{'num-times-accessed':0}};A['bootstrapping-log']={'log-entry':[]}
	if _C in A and A[_C].startswith(_F):A[_C]=sha256_crypt.using(rounds=1000).hash(A[_C][3:])
async def _handle_device_act_code_changed(watched_node_path,jsob,jsob_data_path,nvh):
	A=jsob;assert type(A)==dict
	if jsob_data_path==_B:assert _E in A;B=A[_E]
	else:assert _D in A;B=A[_D]
	if _C in B and B[_C].startswith(_F):B[_C]=sha256_crypt.using(rounds=1000).hash(B[_C][3:])
async def _handle_device_subtree_changed(watched_node_path,jsob,jsob_data_path,nvh):raise NotImplementedError(_G)
async def _handle_device_somehow_changed(watched_node_path,jsob,jsob_data_path,nvh):raise NotImplementedError(_G)
async def _handle_device_deleted(data_path,nvh):0
def _check_expirations(nvh):0


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_U='wn-sztpd-x:'
_T='Unable to parse "input" JSON document: '
_S='malformed-message'
_R='wn-sztpd-x'
_Q='/wn-sztpd-x:tenants/tenant=[^ ]*'
_P='/wn-sztpd-x:tenants/tenant=[^/]*/'
_O='Top node names must begin with the "wn-sztpd-1" prefix.'
_N='application'
_M='wn-sztpd-1'
_L='name'
_K='wn-sztpd-1:'
_J='/wn-sztpd-x:tenants/tenant/0/'
_I='Non-root data_paths must begin with "/wn-sztpd-1:".'
_H='wn-sztpd-x:tenant'
_G='invalid-value'
_F='protocol'
_E='/wn-sztpd-x:tenants/tenant='
_D=':'
_C='/wn-sztpd-1:'
_B=None
_A='/'
import re,json,datetime,basicauth
from aiohttp import web
from passlib.hash import sha256_crypt
from .  import yl
from .  import dal
from .  import utils
from .rcsvr import RestconfServer
from .handler import RouteHandler
class TenantViewHandler(RouteHandler):
	def __init__(A,native):A.native=native
	async def _check_auth(C,request):
		K='access-denied';J='comment';I='failure';H='outcome';E=request;A={};A['timestamp']=datetime.datetime.utcnow();A['source-ip']=E.remote;A['source-proxies']=list(E.forwarded);A['host']=E.host;A['method']=E.method;A['path']=E.path;L=E.headers.get('AUTHORIZATION')
		if L is _B:A[H]=I;A[J]='No authorization specified in the HTTP header.';await C.native._insert_audit_log_entry(_B,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D);return B
		F,O=basicauth.decode(L);R=_B
		try:G=await C.native.dal.get_tenant_name_for_admin(F)
		except dal.NodeNotFound as S:A[H]=I;A[J]='Unknown admin: '+F;await C.native._insert_audit_log_entry(_B,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D);return B
		if G==_B:A[H]=I;A[J]='Host-level admins cannot use tenant interface ('+F+').';await C.native._insert_audit_log_entry(_B,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D);return B
		P=_A+C.native.dal.app_ns+':tenants/tenant='+G+'/admin-accounts/admin-account='+F+'/password';Q=await C.native.dal.handle_get_config_request(P);M=Q[C.native.dal.app_ns+':password'];assert M.startswith('$5$')
		if not sha256_crypt.verify(O,M):A[H]=I;A[J]='Password mismatch for admin '+F;await C.native._insert_audit_log_entry(G,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D);return B
		A[H]='success';await C.native._insert_audit_log_entry(G,A);N=web.Response(status=200);N.text=G;return N
	async def handle_get_opstate_request(F,request):
		E=request;M=E.path;A=M[RestconfServer.len_prefix_operational:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=await F._check_auth(E)
		if B.status==401:return B
		J=B.text;B=await F.native.check_headers(E)
		if B!=_B:return B
		if A=='/ietf-yang-library:yang-library':D=web.Response(status=200);D.content_type='application/yang-data+json';D.text=getattr(yl,'nbi_x_tenant')();return D
		assert A==_A or A.startswith(_C)
		if A==_A:K=_E+J
		else:
			if not A.startswith(_C):D=web.Response(status=400);N=utils.gen_rc_errors(_F,_G,error_message=_I);D.text=json.dumps(N);return D
			Q,L=A.split(_D,1);assert L!=_B;K=_E+J+_A+L
		B,C=await F.native.handle_get_opstate_request_lower_half(K,E.query)
		if C!=_B:
			assert B.status==200;G={}
			if A==_A:
				for H in C[_H][0].keys():
					if H==_L:continue
					G[_K+H]=C[_H][0][H]
			else:I=next(iter(C));assert I.count(_D)==1;O,P=I.split(_D);assert O==_R;assert type(C)==dict;assert len(C)==1;G[_K+P]=C[I]
			B.text=json.dumps(G,indent=2)
		return B
	async def handle_get_config_request(E,request):
		D=request;M=D.path;A=M[RestconfServer.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=await E._check_auth(D)
		if B.status==401:return B
		I=B.text;B=await E.native.check_headers(D)
		if B!=_B:return B
		assert A==_A or A.startswith(_C)
		if A==_A:J=_E+I
		else:
			if not A.startswith(_C):K=web.Response(status=400);N=utils.gen_rc_errors(_F,_G,error_message=_I);K.text=json.dumps(N);return K
			Q,L=A.split(_D,1);assert L!=_B;J=_E+I+_A+L
		B,C=await E.native.handle_get_config_request_lower_half(J,D.query)
		if C!=_B:
			assert B.status==200;F={}
			if A==_A:
				for G in C[_H][0].keys():
					if G==_L:continue
					F[_K+G]=C[_H][0][G]
			else:H=next(iter(C));assert H.count(_D)==1;O,P=H.split(_D);assert O==_R;assert type(C)==dict;assert len(C)==1;F[_K+P]=C[H]
			B.text=json.dumps(F,indent=2)
		return B
	async def handle_post_config_request(G,request):
		D=request;L=D.path;B=L[RestconfServer.len_prefix_running:]
		if B=='':B=_A
		elif B!=_A and B[-1]==_A:B=B[:-1]
		C=await G._check_auth(D)
		if C.status==401:return C
		I=C.text;C=await G.native.check_headers(D)
		if C!=_B:return C
		if B==_A:J=_E+I
		else:
			if not B.startswith(_C):A=web.Response(status=400);E=utils.gen_rc_errors(_F,_G,error_message=_I);A.text=json.dumps(E);return A
			Q,K=B.split(_D,1);assert K!=_B;J=_E+I+_A+K
		try:F=await D.json()
		except json.decoder.JSONDecodeError as M:A=web.Response(status=400);E=utils.gen_rc_errors(_F,_S,error_message=_T+str(M));A.text=json.dumps(E);return A
		assert type(F)==dict;assert len(F)==1;H=next(iter(F));assert H.count(_D)==1;N,O=H.split(_D)
		if N!=_M:A=web.Response(status=400);E=utils.gen_rc_errors(_N,_G,error_message=_O);A.text=json.dumps(E);return A
		P={_U+O:F[H]};A=await G.native.handle_post_config_request_lower_half(J,D.query,P)
		if A.status!=201:
			if'/wn-sztpdex:tenants/tenant/0/'in A.text:A.text=A.text.replace(_J,_C)
			elif _E in A.text:A.text=re.sub(_P,_C,A.text);A.text=re.sub(_Q,_C,A.text)
		return A
	async def handle_put_config_request(G,request):
		F=request;N=F.path;B=N[RestconfServer.len_prefix_running:]
		if B=='':B=_A
		elif B!=_A and B[-1]==_A:B=B[:-1]
		D=await G._check_auth(F)
		if D.status==401:return D
		H=D.text;D=await G.native.check_headers(F)
		if D!=_B:return D
		if B==_A:L=_E+H
		else:
			if not B.startswith(_C):A=web.Response(status=400);C=utils.gen_rc_errors(_F,_G,error_message=_I);A.text=json.dumps(C);return A
			T,M=B.split(_D,1);assert M!=_B;L=_E+H+_A+M
		try:E=await F.json()
		except json.decoder.JSONDecodeError as O:A=web.Response(status=400);C=utils.gen_rc_errors(_F,_S,error_message=_T+str(O));A.text=json.dumps(C);return A
		if B==_A:
			I={_H:[{_L:H}]}
			for J in E.keys():
				assert J.count(_D)==1;P,Q=J.split(_D)
				if P!=_M:A=web.Response(status=400);C=utils.gen_rc_errors(_N,_G,error_message=_O);A.text=json.dumps(C);return A
				I[_H][0][Q]=E[J]
		else:
			assert type(E)==dict;assert len(E)==1;K=next(iter(E));assert K.count(_D)==1;R,S=K.split(_D)
			if R!=_M:A=web.Response(status=400);C=utils.gen_rc_errors(_N,_G,error_message=_O);A.text=json.dumps(C);return A
			I={_U+S:E[K]}
		A=await G.native.handle_put_config_request_lower_half(L,F.query,I)
		if A.status!=204:
			if _J in A.text:A.text=A.text.replace(_J,_C)
			elif _E in A.text:A.text=re.sub(_P,_C,A.text);A.text=re.sub(_Q,_C,A.text)
		return A
	async def handle_delete_config_request(D,request):
		E=request;I=E.path;B=I[RestconfServer.len_prefix_running:]
		if B=='':B=_A
		elif B!=_A and B[-1]==_A:B=B[:-1]
		C=await D._check_auth(E)
		if C.status==401:return C
		F=C.text;C=await D.native.check_headers(E)
		if C!=_B:return C
		if B==_A:G=_E+F
		else:
			if not B.startswith(_C):A=web.Response(status=400);J=utils.gen_rc_errors(_F,_G,error_message=_I);A.text=json.dumps(J);return A
			K,H=B.split(_D,1);assert H!=_B;G=_E+F+_A+H
		A=await D.native.handle_delete_config_request_lower_half(G)
		if A.status!=204:
			if _J in A.text:A.text=A.text.replace(_J,_C)
			elif _E in A.text:A.text=re.sub(_P,_C,A.text);A.text=re.sub(_Q,_C,A.text)
		return A
	async def handle_action_request(A,request):0
	async def handle_rpc_request(A,request):0


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
from abc import ABCMeta,abstractmethod
from aiohttp import web
class RouteHandler(metaclass=ABCMeta):
	@abstractmethod
	async def handle_get_opstate_request(self,request):0
	@abstractmethod
	async def handle_get_config_request(self,request):0
	@abstractmethod
	async def handle_post_config_request(self,request):0
	@abstractmethod
	async def handle_put_config_request(self,request):0
	@abstractmethod
	async def handle_delete_config_request(self,request):0
	@abstractmethod
	async def handle_action_request(self,request):0
	@abstractmethod
	async def handle_rpc_request(self,request):0


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_A='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2020-08-27",\n        "feature": [\n          "device-ownership-verification"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
def nbi_0():return bss_common+bss_0_vZ
def nbi_1():return bss_common+bss_1_vZ
def nbi_x():return bss_common+bss_x_vZ
def nbi_x_tenant():return bss_common+bss_x_tenant_vZ
def sbi_rfc8572():return sbi_rfc8572_yl_str
def sbi_legacy():return'FIXME'
sbi_rfc8572_yl_str='\n{\n  "ietf-yang-library:modules-state": {\n    "module-set-id": "TBD",\n    "module": [\n      {\n        "name": "ietf-yang-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-inet-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-netconf-acm",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-netconf-acm",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-datastores",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-datastores",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-yang-library",\n        "revision": "2019-01-04",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-sztp-bootstrap-server",\n        "revision": "2019-04-30",\n        "feature": [\n          "onboarding-server"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-yang-structure-ext",\n        "revision": "2019-03-07",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-crypto-types",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-crypto-types",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-sztp-csr",\n        "revision": "2020-06-02",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-csr",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}'
bss_common='\n{\n  "ietf-yang-library:modules-state": {\n    "module-set-id": "TBD",\n    "module": [\n      {\n        "name": "ietf-yang-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-inet-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-datastores",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-datastores",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-yang-library",\n        "revision": "2019-01-04",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",\n        "conformance-type": "import"\n      },\n      {\n        "name": "iana-crypt-hash",\n        "revision": "2014-08-06",\n        "namespace": "urn:ietf:params:xml:ns:yang:iana-crypt-hash",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-x509-cert-to-name",\n        "revision": "2014-12-10",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-x509-cert-to-name",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-restconf",\n        "revision": "2017-01-26",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-restconf",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-netconf-acm",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-netconf-acm",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-sztp-conveyed-info",\n        "revision": "2019-04-30",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-conveyed-info",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-crypto-types",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-crypto-types",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-truststore",\n        "revision": "2020-08-27",\n        "feature": [\n          "certificates"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-truststore",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-keystore",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-keystore",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tcp-common",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tcp-common",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tcp-client",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tcp-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tcp-server",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tcp-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tls-common",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tls-common",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tls-client",\n        "revision": "2020-08-27",\n        "feature": [\n          "x509-certificate-auth",\n          "client-auth-config-supported"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tls-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tls-server",\n        "revision": "2020-08-27",\n        "feature": [\n          "x509-certificate-auth",\n          "client-auth-config-supported"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tls-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-http-client",\n        "revision": "2020-08-27",\n        "feature": [\n          "basic-auth"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-http-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-http-server",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-http-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-restconf-client",\n        "revision": "2020-08-27",\n        "feature": [\n          "https-initiate"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-restconf-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-restconf-server",\n        "revision": "2020-08-27",\n        "feature": [\n          "http-listen",\n          "https-listen"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-restconf-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-x509-c2n",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/wnc2n",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-app-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/app-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-app",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/wnapp",\n        "feature": [\n          "webhook-based-callouts",\n          "http-based-webhooks"\n        ],\n        "conformance-type": "import"\n      },\n'
bss_0_vZ='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_1_vZ=_A
bss_x_vZ='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2020-08-27",\n        "feature": [\n          "device-ownership-verification"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-x",\n        "revision": "2020-08-27",\n        "feature": [\n          "device-ownership-verification"\n        ],\n        "namespace": "https://watsen.net/sztpd-x",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_x_tenant_v1='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2020-08-27",\n        "feature": [\n          "system-level-administration-disabled"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_x_tenant_vZ=_A


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_f='should check to see if an alarm can be cleared...'
_e='admin-account'
_d='$0$'
_c='%Y-%m-%dT%H:%M:%SZ'
_b='password-last-modified'
_a='module'
_Z='sztpd.plugins.'
_Y="why wasn't this assertion caught by val? "
_X='operation-failed'
_W='data-exists'
_V='Unable to parse "input" JSON document: '
_U='malformed-message'
_T='method'
_S='\\g<1>'
_R='.*plugins/plugin=([^/]*).*'
_Q='missing-attribute'
_P='function'
_O='need to implement this code'
_N='SZTPD_MODE'
_M='application/yang-data+json'
_L='name'
_K='plugin'
_J='operation-not-supported'
_I='functions'
_H='password'
_G='unknown-element'
_F='sleep'
_E='invalid-value'
_D='application'
_C='protocol'
_B=None
_A='/'
import os,re,sys,json,base64,signal,asyncio,yangson,datetime,basicauth,importlib,pkg_resources
from enum import Enum
from aiohttp import web
from enum import IntFlag
from pyasn1.type import univ
from fifolock import FifoLock
from passlib.hash import sha256_crypt
from .dal import DataAccessLayer
from .val import ValidationLayer
from .rcsvr import RestconfServer
from .handler import RouteHandler
from .  import dal
from .  import val
from .  import utils
class RefAction(IntFlag):ADDED=1;REMOVED=2
class TimeUnit(Enum):Days=2;Hours=1;Minutes=0
class Period:
	def __init__(A,amount,units):A.amount=amount;A.units=units
class PluginNotFound(Exception):0
class PluginSyntaxError(Exception):0
class FunctionNotFound(Exception):0
class FunctionNotCallable(Exception):0
class Read(asyncio.Future):
	@staticmethod
	def is_compatible(holds):return not holds[Write]
class Write(asyncio.Future):
	@staticmethod
	def is_compatible(holds):A=holds;return not A[Read]and not A[Write]
class NativeViewHandler(RouteHandler):
	len_prefix_running=RestconfServer.len_prefix_running;len_prefix_operational=RestconfServer.len_prefix_operational;len_prefix_operations=RestconfServer.len_prefix_operations
	def __init__(A,_dal,_mode,_loop):
		O=':preferences/system/plugins/plugin/functions/function';N=':preferences/system/plugins/plugin';M=':tenants/tenant/admin-accounts/admin-account/password';L=':admin-accounts/admin-account/password';K=':plugins';A.dal=_dal;A.mode=_mode;A.loop=_loop;A.fifolock=FifoLock();A.create_callbacks={};A.change_callbacks={};A.delete_callbacks={};A.subtree_change_callbacks={};A.somehow_change_callbacks={};A.leafref_callbacks={};A.periodic_callbacks={};A.onetime_callbacks={};A.plugins={};B=A.dal.handle_get_opstate_request('/ietf-yang-library:yang-library');F=A.loop.run_until_complete(B);G=pkg_resources.resource_filename('sztpd','yang/');A.dm=yangson.DataModel(json.dumps(F),[G]);A.val=ValidationLayer(A.dm,A.dal);B=A.dal.handle_get_opstate_request(_A+A.dal.app_ns+':preferences/system/plugins')
		try:D=A.loop.run_until_complete(B)
		except dal.NodeNotFound:pass
		else:
			if _K in D[A.dal.app_ns+K]:
				for C in D[A.dal.app_ns+K][_K]:
					H=C[_L];B=_handle_plugin_created('',{_K:C},'',A,_B);A.loop.run_until_complete(B)
					if _I in C:
						for E in C[_I][_P]:P=E[_L];I='FOO/plugins/plugin='+H+'/BAR';B=_handle_function_created('',{_P:E},I,A,_B);A.loop.run_until_complete(B)
		A.register_create_callback(_A+A.dal.app_ns+L,_handle_admin_passwd_created);A.register_change_callback(_A+A.dal.app_ns+L,_handle_admin_passwd_changed)
		if A.mode=='x':A.register_create_callback(_A+A.dal.app_ns+M,_handle_admin_passwd_created);A.register_change_callback(_A+A.dal.app_ns+M,_handle_admin_passwd_changed)
		A.register_create_callback(_A+A.dal.app_ns+':tenants/tenant',_handle_tenant_created);A.register_create_callback(_A+A.dal.app_ns+N,_handle_plugin_created);A.register_delete_callback(_A+A.dal.app_ns+N,_handle_plugin_deleted);A.register_create_callback(_A+A.dal.app_ns+O,_handle_function_created);A.register_delete_callback(_A+A.dal.app_ns+O,_handle_function_deleted);A.register_change_callback(_A+A.dal.app_ns+':transport/listen',_handle_transport_changed);A.register_delete_callback(_A+A.dal.app_ns+':transport',_handle_transport_delete);A.register_periodic_callback(Period(24,TimeUnit.Hours),datetime.datetime(2000,1,1,0),_check_expirations)
		for J in A.dal.ref_stat_collectors:A.register_create_callback(J.replace('/reference-statistics',''),_handle_ref_stat_parent_created)
	def register_create_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.create_callbacks:A.create_callbacks[B]=[C]
		else:A.create_callbacks[B].append(C)
	def register_change_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.change_callbacks:A.change_callbacks[B]=[C]
		else:A.change_callbacks[B].append(C)
	def register_subtree_change_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.subtree_change_callbacks:A.subtree_change_callbacks[B]=[C]
		else:A.subtree_change_callbacks[B].append(C)
	def register_somehow_change_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.somehow_change_callbacks:A.somehow_change_callbacks[B]=[C]
		else:A.somehow_change_callbacks[B].append(C)
	def register_delete_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.delete_callbacks:A.delete_callbacks[B]=[C]
		else:A.delete_callbacks[B].append(C)
	def register_onetime_callback(A,timestamp,callback,opaque):
		B=callback
		if schema_path not in A.onetime_callbacks:A.onetime_callbacks[schema_path]=[B]
		else:A.onetime_callbacks[schema_path].append(B)
	def register_periodic_callback(A,period,anchor,callback):0
	def register_leafref_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.leafref_callbacks:A.leafref_callbacks[B]=[C]
		else:A.leafref_callbacks[B].append(C)
	async def _insert_audit_log_entry(A,tenant_name,audit_log_entry):
		C=audit_log_entry;B=tenant_name
		if C[_T]in{'GET','HEAD'}:return
		if B==_B:D=_A+A.dal.app_ns+':audit-log'
		else:F=A.dal.opaque();assert F=='x';D=_A+A.dal.app_ns+':tenants/tenant='+B+'/audit-log'
		E={};E[A.dal.app_ns+':log-entry']=C;await A.dal.handle_post_opstate_request(D,E)
	async def _check_auth(B,request,data_path):
		S='No authorization required for fresh installs.';R=':admin-accounts/admin-account';N='access-denied';M='failure';L='success';G='comment';F='outcome';D=request;A={};A['timestamp']=datetime.datetime.utcnow();A['source-ip']=D.remote;A['source-proxies']=list(D.forwarded);A['host']=D.host;A[_T]=D.method;A['path']=D.path;J=D.headers.get('AUTHORIZATION')
		if J is _B:
			H=await B.dal.num_elements_in_list(_A+B.dal.app_ns+R)
			if H==0:A[F]=L;A[G]=S;await B._insert_audit_log_entry(_B,A);return web.Response(status=200)
			A[F]=M;A[G]='No authorization specified in the HTTP header.';await B._insert_audit_log_entry(_B,A);C=web.Response(status=401);E=utils.gen_rc_errors(_C,N);C.text=json.dumps(E);return C
		I,O=basicauth.decode(J);P=_A+B.dal.app_ns+':admin-accounts/admin-account='+I+'/password'
		try:Q=await B.dal.handle_get_config_request(P)
		except dal.NodeNotFound as T:
			H=await B.dal.num_elements_in_list(_A+B.dal.app_ns+R)
			if H==0:A[F]=L;A[G]=S;await B._insert_audit_log_entry(_B,A);return web.Response(status=200)
			A[F]=M;A[G]='Unknown admin: '+I;await B._insert_audit_log_entry(_B,A);C=web.Response(status=401);E=utils.gen_rc_errors(_C,N);C.text=json.dumps(E);return C
		K=Q[B.dal.app_ns+':password'];assert K.startswith('$5$')
		if not sha256_crypt.verify(O,K):A[F]=M;A[G]='Password mismatch for admin '+I;await B._insert_audit_log_entry(_B,A);C=web.Response(status=401);E=utils.gen_rc_errors(_C,N);C.text=json.dumps(E);return C
		A[F]=L;await B._insert_audit_log_entry(_B,A);return web.Response(status=200)
	async def check_headers(F,request):
		E='Accept';D='Content-Type';B=request
		if any((B.method==A for A in('PUT','POST','PATCH'))):
			if B.body_exists:
				if D not in B.headers:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_Q,error_message='"'+B.method+'" request missing the "Content-Type" header (RFC 8040, 5.2).');A.text=json.dumps(C);return A
				if B.headers[D]!=_M:A=web.Response(status=415);C=utils.gen_rc_errors(_C,_J,error_message='Content-Type, when specified, must be "application/yang-data+json".');A.text=json.dumps(C);return A
		if E in B.headers:
			if not any((B.headers[E]==A for A in('*/*','application/*',_M))):A=web.Response(status=406);C=utils.gen_rc_errors(_C,_E,error_message='The "Accept" type, when set, must be "*/*", "application/*", or "application/yang-data+json".');A.text=json.dumps(C);return A
	async def handle_get_opstate_request(C,request):
		D=request;A=D.path[C.len_prefix_operational:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=await C._check_auth(D,A)
		if B.status==401:return B
		B=await C.check_headers(D)
		if B!=_B:return B
		B,E=await C.handle_get_opstate_request_lower_half(A,D.query)
		if E!=_B:B.text=json.dumps(E,indent=4)
		return B
	async def handle_get_opstate_request_lower_half(D,data_path,query_dict):
		E=query_dict
		async with D.fifolock(Read):
			if os.environ.get(_N)and _F in E:await asyncio.sleep(int(E[_F]))
			try:F=await D.dal.handle_get_opstate_request(data_path)
			except dal.NodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_C,_G,error_message=str(B));A.text=json.dumps(C);return A,_B
			except NotImplementedError as B:A=web.Response(status=501);C=utils.gen_rc_errors(_D,_J,error_message=str(B));A.text=json.dumps(C);return A,_B
			A=web.Response(status=200);A.content_type=_M;return A,F
	async def handle_get_config_request(C,request):
		D=request;A=D.path[C.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=await C._check_auth(D,A)
		if B.status==401:return B
		B=await C.check_headers(D)
		if B!=_B:return B
		B,E=await C.handle_get_config_request_lower_half(A,D.query)
		if E!=_B:B.text=json.dumps(E,indent=4)
		return B
	async def handle_get_config_request_lower_half(D,data_path,query_dict):
		F=data_path;E=query_dict
		async with D.fifolock(Read):
			try:await D.val.handle_get_config_request(F,E)
			except val.InvalidDataPath as B:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_E,error_message=str(B));A.text=json.dumps(C);return A,_B
			except val.NonexistentSchemaNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_E,error_message=str(B));A.text=json.dumps(C);return A,_B
			except val.NodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_C,_G,error_message=str(B));A.text=json.dumps(C);return A,_B
			if os.environ.get(_N)and _F in E:await asyncio.sleep(int(E[_F]))
			try:G=await D.dal.handle_get_config_request(F)
			except dal.NodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_C,_G,error_message=str(B));A.text=json.dumps(C);return A,_B
			A=web.Response(status=200);A.content_type=_M;return A,G
	async def handle_post_config_request(D,request):
		B=request;A=B.path[D.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		C=await D._check_auth(B,A)
		if C.status==401:return C
		C=await D.check_headers(B)
		if C!=_B:return C
		try:F=await B.json()
		except json.decoder.JSONDecodeError as G:E=web.Response(status=400);H=utils.gen_rc_errors(_C,_U,error_message=_V+str(G));E.text=json.dumps(H);return E
		return await D.handle_post_config_request_lower_half(A,B.query,F)
	async def handle_post_config_request_lower_half(D,data_path,query_dict,request_body):
		G=request_body;F=data_path;E=query_dict
		async with D.fifolock(Write):
			try:await D.val.handle_post_config_request(F,E,G)
			except (val.InvalidInputDocument,val.UnrecognizedQueryParameter)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_E,error_message=str(B));A.text=json.dumps(C);return A
			except val.MissingQueryParameter as B:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_Q,error_message=str(B));A.text=json.dumps(C);return A
			except val.NonexistentSchemaNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_E,error_message=str(B));A.text=json.dumps(C);return A
			except val.ValidationFailed as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_E,error_message=str(B));A.text=json.dumps(C);return A
			except val.ParentNodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_C,_G,error_message=str(B));A.text=json.dumps(C);return A
			except val.UnrecognizedInputNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_G,error_message=str(B));A.text=json.dumps(C);return A
			except NotImplementedError as B:A=web.Response(status=501);C=utils.gen_rc_errors(_D,_J,error_message=str(B));A.text=json.dumps(C);return A
			except val.NodeAlreadyExists as B:A=web.Response(status=409);C=utils.gen_rc_errors(_D,_W,error_message=str(B));A.text=json.dumps(C);return A
			if os.environ.get(_N)and _F in E:await asyncio.sleep(int(E[_F]))
			try:await D.dal.handle_post_config_request(F,G,D.create_callbacks,D.change_callbacks,D)
			except dal.CreateCallbackFailed as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_X,error_message=str(B));A.text=json.dumps(C);return A
			except Exception as B:raise Exception(_Y+str(B))
			D.val.inst=D.val.inst2;D.val.inst2=_B;await D.shared_post_commit_logic();return web.Response(status=201)
	async def handle_put_config_request(D,request):
		B=request;A=B.path[D.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		C=await D._check_auth(B,A)
		if C.status==401:return C
		C=await D.check_headers(B)
		if C!=_B:return C
		try:F=await B.json()
		except json.decoder.JSONDecodeError as G:E=web.Response(status=400);H=utils.gen_rc_errors(_C,_U,error_message=_V+str(G));E.text=json.dumps(H);return E
		return await D.handle_put_config_request_lower_half(A,B.query,F)
	async def handle_put_config_request_lower_half(D,data_path,query_dict,request_body):
		G=request_body;F=data_path;E=query_dict
		async with D.fifolock(Write):
			try:await D.val.handle_put_config_request(F,E,G)
			except val.InvalidDataPath as B:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_E,error_message=str(B));A.text=json.dumps(C);return A
			except val.ParentNodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_C,_G,error_message=str(B));A.text=json.dumps(C);return A
			except val.UnrecognizedInputNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_G,error_message=str(B));A.text=json.dumps(C);return A
			except (val.NonexistentSchemaNode,val.ValidationFailed)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_E,error_message=str(B));A.text=json.dumps(C);return A
			except (val.InvalidInputDocument,val.UnrecognizedQueryParameter)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_E,error_message=str(B));A.text=json.dumps(C);return A
			except val.MissingQueryParameter as B:A=web.Response(status=400);C=utils.gen_rc_errors(_C,_Q,error_message=str(B));A.text=json.dumps(C);return A
			except val.NodeAlreadyExists as B:A=web.Response(status=409);C=utils.gen_rc_errors(_D,_W,error_message=str(B));A.text=json.dumps(C);return A
			except NotImplementedError as B:A=web.Response(status=501);C=utils.gen_rc_errors(_D,_J,error_message=str(B));A.text=json.dumps(C);return A
			if os.environ.get(_N)and _F in E:await asyncio.sleep(int(E[_F]))
			try:await D.dal.handle_put_config_request(F,G,D.create_callbacks,D.change_callbacks,D.delete_callbacks,D)
			except dal.CreateCallbackFailed as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_X,error_message=str(B));A.text=json.dumps(C);return A
			except (PluginNotFound,PluginSyntaxError,FunctionNotFound,FunctionNotCallable)as B:A=web.Response(status=501);C=utils.gen_rc_errors(_D,_J,error_message=str(B));A.text=json.dumps(C);return A
			except Exception as B:raise Exception("why wasn't this assertion caught by val? (assuming it's a YANG validation thing)"+str(B))
			D.val.inst=D.val.inst2;D.val.inst2=_B;await D.shared_post_commit_logic();return web.Response(status=204)
	async def handle_delete_config_request(C,request):
		D=request;A=D.path[C.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=await C._check_auth(D,A)
		if B.status==401:return B
		B=await C.check_headers(D)
		if B!=_B:return B
		return await C.handle_delete_config_request_lower_half(A)
	async def handle_delete_config_request_lower_half(A,data_path):
		E=data_path
		async with A.fifolock(Write):
			try:await A.val.handle_delete_config_request(E)
			except val.NonexistentSchemaNode as C:B=web.Response(status=400);D=utils.gen_rc_errors(_D,_E,error_message=str(C));B.text=json.dumps(D);return B
			except val.NodeNotFound as C:B=web.Response(status=404);D=utils.gen_rc_errors(_C,_G,error_message=str(C));B.text=json.dumps(D);return B
			except val.ValidationFailed as C:B=web.Response(status=400);D=utils.gen_rc_errors(_D,_E,error_message=str(C));B.text=json.dumps(D);return B
			try:await A.dal.handle_delete_config_request(E,A.delete_callbacks,A.change_callbacks,A)
			except Exception as C:raise Exception(_Y+str(C))
			A.val.inst=A.val.inst2;A.val.inst2=_B;await A.shared_post_commit_logic();return web.Response(status=204)
	async def shared_post_commit_logic(A):0
	async def handle_action_request(A,request):0
	async def handle_rpc_request(A,request):raise NotImplementedError('Native needs an RPC handler?  - client accessible!')
	def _handle_generate_symmetric_key_action(A,data_path,action_input):raise NotImplementedError(_O)
	def _handle_generate_asymmetric_key_action(A,data_path,action_input):raise NotImplementedError(_O)
	def _handle_resend_activation_email_action(A,data_path,action_input):raise NotImplementedError(_O)
	def _handle_generate_certificate_signing_request_action(A,data_path,action_input):raise NotImplementedError(_O)
async def _handle_tenant_created(watched_node_path,jsob,jsob_data_path,obj,conn):jsob['tenant']['audit-log']={}
async def _handle_transport_changed(watched_node_path,jsob,jsob_data_path,obj):os.kill(os.getpid(),signal.SIGHUP)
async def _handle_transport_delete(watched_node_path,opaque):raise NotImplementedError('Deleting the /transport node itself cannot be constrained by YANG.')
async def _handle_plugin_created(watched_node_path,jsob,jsob_data_path,opaque,conn):
	B=opaque;A=jsob[_K][_L];C=_Z+A
	if A in B.plugins:E=sys.modules[C];del sys.modules[C];del E;del B.plugins[A]
	try:F=importlib.import_module(C)
	except ModuleNotFoundError as D:raise PluginNotFound(str(D))
	except SyntaxError as D:raise PluginSyntaxError('SyntaxError: '+str(D))
	B.plugins[A]={_a:F,_I:{}}
async def _handle_plugin_deleted(watched_node_path,opaque):C=opaque;A=re.sub(_R,_S,watched_node_path);B=_Z+A;D=sys.modules[B];del sys.modules[B];del D;del C.plugins[A]
async def _handle_function_created(watched_node_path,jsob,jsob_data_path,opaque,conn):
	B=opaque;C=re.sub(_R,_S,jsob_data_path);A=jsob[_P][_L]
	try:D=getattr(B.plugins[C][_a],A)
	except AttributeError as E:raise FunctionNotFound(str(E))
	if not callable(D):raise FunctionNotCallable("The plugin function name '"+A+"' is not callable.")
	B.plugins[C][_I][A]=D
async def _handle_function_deleted(watched_node_path,opaque):A=watched_node_path;B=opaque;C=re.sub(_R,_S,A);D=A.rsplit('=',1)[1];del B.plugins[C][_I][D]
async def _handle_admin_passwd_created(watched_node_path,jsob,jsob_data_path,obj,conn):
	A=jsob
	def B(item):
		A=item;A[_b]=datetime.datetime.utcnow().strftime(_c)
		if _H in A and A[_H].startswith(_d):A[_H]=sha256_crypt.using(rounds=1000).hash(A[_H][3:])
	if type(A)==dict:B(A[_e])
	else:
		assert False;assert type(A)==list
		for C in A:assert type(C)==dict;B(C)
async def _handle_admin_passwd_changed(watched_node_path,json,jsob_data_path,obj):
	def A(item):
		A=item;A[_b]=datetime.datetime.utcnow().strftime(_c)
		if _H in A and A[_H].startswith(_d):A[_H]=sha256_crypt.using(rounds=1000).hash(A[_H][3:])
		else:0
	assert json!=_B;assert jsob_data_path!=_B;A(json[_e])
async def _handle_ref_stat_parent_created(watched_node_path,jsob,jsob_data_path,obj,conn):
	A=jsob;assert watched_node_path==jsob_data_path
	def B(item):item['reference-statistics']={'reference-count':0,'last-referenced':'never'}
	if type(A)==dict:D=next(iter(A));B(A[D])
	else:
		raise NotImplementedError('dead code?');assert type(A)==list
		for C in A:assert type(C)==dict;B(C)
def _handle_ref_stats_changed(leafrefed_node_data_path,obj):raise NotImplementedError('_handle_ref_stats_changed tested?')
def _handle_lingering_unreferenced_node_change(watched_node_path,obj):raise NotImplementedError(_f)
def _handle_expiring_certificate_change(watched_node_path,obj):raise NotImplementedError(_f)
async def _check_expirations(nvh):0


import io
import os
import sys
import setuptools
#from glob import glob
from setuptools.command.sdist import sdist

# a combination of:
#  -  https://realpython.com/pypi-publish-python-package
#  -  https://github.com/kennethreitz/setup.py/blob/master/setup.py

NAME = "sztpd"
HERE = os.path.abspath(os.path.dirname(__file__))
#HERE = pathlib.Path(__file__).parent

# Import the README.md and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
with io.open(os.path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(HERE, 'src', NAME, '__version__.py')) as f:
    exec(f.read(), about)
with open(os.path.join(HERE, 'README.md')) as f:
    about['__long__'] = f.read()
#README = (HERE / "README.md").read_text()

#def read_text(file_name: str):
#    return open(os.path.join(base_path, file_name)).read()


class CustomSdistCommand(sdist):
    def run(self):
        import git
        import subprocess
        import python_minifier
        from os import listdir
        #from pyang.translators import yang as yang_translator

        # ensure all files checked in
        repo = git.Repo("./")
        if len(repo.index.diff(None)) > 0:
            print("\nPlease commit files and try again.")
            return

        # minimize the .py files
        #FIXME: any way to remove all the comments in the YANG files?
        for fn in listdir("src/sztpd"):
            if fn.endswith('.py'):
                print("Processing " + fn + "...")
        
                with open("src/sztpd/"+fn, mode='r') as f:
                    minified = python_minifier.minify(f.read(), filename=fn, remove_literal_statements=True)

                with open("src/sztpd/"+fn, mode='w') as f:
                    f.write("# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.\n\n")
                    f.write(minified)


        # strip comments out of YANG files
        for fn in listdir("src/sztpd/yang"):
            if fn.startswith('wn-'):
                print("Processing " + fn + "...")
                subprocess.run(["pyang", "-p", "src/sztpd/yang/", "-f", "yang", "--yang-remove-comments", "src/sztpd/yang/"+fn, "-o", "src/sztpd/yang/"+fn])

#pyang -f yang --yang-remove-comments $i -o $i

# OLD PYANG module-based approach (SHOULD FIX!)
#        # strip comments out of YANG files
#        yang_translator = YANGPlugin()
#        yang_translator.add_opts(["--yang-remove-comments"])
#        ctx = 
#
#        for filename in filenames:
#            try:
#                fd = io.open(filename, "r", encoding="utf-8")
#                text = fd.read()
#                module = ctx.add_module(filename, text)
#
#            text = sys.stdin.read()
#            module = ctx.add_module('<stdin>', text)
#
#          fd = open...
#          yang_translator.emit(ctx?, modules?, fd?)


        # let normal sdist process run
        sdist.run(self)

        # restore Git files to original state
        head = repo.heads[0]
        head.checkout(force=True)


# Where the magic happens:
setuptools.setup(
    name=NAME,
    description="Secure Zero Touch Provisioning Daemon (SZTPD)",
    #long_description=long_description,
    long_description=about['__long__'],
    long_description_content_type='text/markdown',
    license='Non Production License',
    #license_file="LICENSE",
    url="https://watsen.net/products/sztpd",
    author="Watsen Networks",
    author_email="info@watsen.net",
    project_urls={
      "Documentation": "https://watsen.net/docs/sztpd",
      "License": "https://watsen.net/sales/non-production-use.html"
      #"Bug Tracker": "https://bugs.example.com/HelloWorld/",
      #"Source Code": "https://code.example.com/HelloWorld/",
    },
    version=about['__version__'],
    python_requires=">=3.7.*, <4",
    package_dir = {"": "src"},
    packages=setuptools.find_packages("src"),
    #py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    #packages=['src/sztpd'],
    #package_dir={'sztpd': 'src/sztpd'},
    package_data={'sztpd': ['LICENSE.txt', 'yang/*.yang']},
    #include_package_data = True,
    keywords = ["IETF", "SZTP", "RFC 8572", "secure", "zerotouch", "zero touch", "provisioning", "ztp"],
    install_requires=['six', 'sqlalchemy', 'sqlalchemy_utils', 'pysqlite3', 'cryptography', 'yangson', 'aiohttp==3.6.2', 'basicauth', 'pyasn1', 'pyasn1-modules', 'pem', 'passlib', 'requests', 'certvalidator', 'fifolock'],
    #extras_require=EXTRAS,
    #setup_requires = ["pytest-runner"],
    #tests_require = ["pytest"],
    entry_points = { "console_scripts": ["sztpd=sztpd.__main__:main"] },
    classifiers = [
      # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
      "Development Status :: 3 - Alpha",
      "Environment :: Console",
      "Intended Audience :: Information Technology",
      "Intended Audience :: System Administrators",
      "Intended Audience :: Telecommunications Industry",
      "License :: Other/Proprietary License",
      "Natural Language :: English",
      "Operating System :: OS Independent",
      "Programming Language :: Python",
      "Programming Language :: Python :: 3",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Topic :: Communications",
      "Topic :: Internet",
      "Topic :: Security"
    ],
    cmdclass={
        'sdist': CustomSdistCommand
    }
)




# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_E='application/yang-data+json'
_D='/ds/ietf-datastores:operational'
_C='/ds/ietf-datastores:running'
_B=None
_A='/'
import os,ssl,json,base64,pyasn1,asyncio,yangson,datetime,tempfile,basicauth
from .  import utils
from aiohttp import web
from .handler import RouteHandler
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5915
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.codec.der.encoder import encode as der_encoder
async def set_server_header(request,response):response.headers['Server']='<redacted>'
class RestconfServer:
	root='/restconf';prefix_running=root+_C;prefix_operational=root+_D;prefix_operations=root+'/operations';len_prefix_running=len(prefix_running);len_prefix_operational=len(prefix_operational);len_prefix_operations=len(prefix_operations)
	def __init__(A,loop,dal,endpoint_config,view_handler,facade_yl=_B):
		A3='client-certs';A2='local-truststore-reference';A1='ca-certs';A0='client-authentication';z='w';y='cert-data';x='ASCII';w=':keystore/asymmetric-keys/asymmetric-key=';v='reference';u='server-identity';t='local-port';s='http';X='tcp-server-parameters';N='certificate';M='tls-server-parameters';L='/ds/ietf-datastores:running{tail:.*}';G=dal;E='https';C=view_handler;B=endpoint_config;A.len_prefix_running=len(A.root+_C);A.len_prefix_operational=len(A.root+_D);A.loop=loop;A.dal=G;A.name=B['name'];A.view_handler=C;A.app=web.Application();A.app.on_response_prepare.append(set_server_header);A.app.router.add_get('/.well-known/host-meta',A.handle_get_host_meta);A.app.router.add_get(A.root,A.handle_get_tippy_top);A.app.router.add_get(A.root+_A,A.handle_get_tippy_top);A.app.router.add_get(A.root+'/yang-library-version',A.handle_get_yang_library_version);A.app.router.add_get(A.root+'/ds/ietf-datastores:operational{tail:.*}',C.handle_get_opstate_request);A.app.router.add_get(A.root+L,C.handle_get_config_request);A.app.router.add_put(A.root+L,C.handle_put_config_request);A.app.router.add_post(A.root+L,C.handle_post_config_request);A.app.router.add_delete(A.root+L,C.handle_delete_config_request);A.app.router.add_post(A.root+'/ds/ietf-datastores:operational/{tail:.*}',C.handle_action_request);A.app.router.add_post(A.root+'/operations/{tail:.*}',C.handle_rpc_request)
		if s in B:F=s
		else:assert E in B;F=E
		A.local_address=B[F][X]['local-address'];A.local_port=os.environ.get('SZTPD_DEFAULT_PORT',8080)
		if t in B[F][X]:A.local_port=B[F][X][t]
		D=_B
		if F==E:
			O=B[E][M][u][N][v]['asymmetric-key'];I=A.dal.handle_get_config_request(_A+A.dal.app_ns+w+O);Y=A.loop.run_until_complete(I);P=Y[A.dal.app_ns+':asymmetric-key'][0]['cleartext-private-key'];Z=base64.b64decode(P);a,A4=der_decoder(Z,asn1Spec=rfc5915.ECPrivateKey());b=der_encoder(a);Q=base64.b64encode(b).decode(x);assert P==Q;c='-----BEGIN EC PRIVATE KEY-----\n'+Q+'\n-----END EC PRIVATE KEY-----\n';d=B[E][M][u][N][v][N];I=A.dal.handle_get_config_request(_A+A.dal.app_ns+w+O+'/certificates/certificate='+d);e=A.loop.run_until_complete(I);f=e[A.dal.app_ns+':certificate'][0][y];g=base64.b64decode(f);h,i=der_decoder(g,asn1Spec=rfc5652.ContentInfo());j=h.getComponentByName('content');k,i=der_decoder(j,asn1Spec=rfc5652.SignedData());R=k.getComponentByName('certificates');S=''
			for l in range(len(R)):m=R[l][0];n=der_encoder(m);o=base64.b64encode(n).decode(x);S+='-----BEGIN CERTIFICATE-----\n'+o+'\n-----END CERTIFICATE-----\n'
			D=ssl.create_default_context(ssl.Purpose.CLIENT_AUTH);D.verify_mode=ssl.CERT_OPTIONAL
			with tempfile.TemporaryDirectory()as T:
				U=T+'key.pem';V=T+'certs.pem'
				with open(U,z)as p:p.write(c)
				with open(V,z)as q:q.write(S)
				D.load_cert_chain(V,U)
			if A0 in B[E][M]:
				H=B[E][M][A0]
				def W(truststore_ref):
					C=G.handle_get_config_request(_A+G.app_ns+':truststore/certificate-bags/certificate-bag='+truststore_ref);D=A.loop.run_until_complete(C);B=[]
					for E in D[G.app_ns+':certificate-bag'][0][N]:F=base64.b64decode(E[y]);H,I=der_decoder(F,asn1Spec=rfc5652.ContentInfo());assert not I;B+=utils.degenerate_cms_obj_to_ders(H)
					return B
				J=[]
				if A1 in H:K=H[A1][A2];J+=W(K)
				if A3 in H:K=H[A3][A2];J+=W(K)
				r=utils.der_dict_to_multipart_pem({'CERTIFICATE':J});D.load_verify_locations(cadata=r)
		if F==E:assert not D is _B
		else:assert D is _B
		A.runner=web.AppRunner(A.app);A.loop.run_until_complete(A.runner.setup());A.site=web.TCPSite(A.runner,host=A.local_address,port=A.local_port,ssl_context=D,reuse_port=True);A.loop.run_until_complete(A.site.start())
	async def handle_get_host_meta(B,request):A=web.Response();A.content_type='application/xrd+xml';A.text='<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">\n  <Link rel="restconf" href="/restconf"/>\n</XRD>';return A
	async def handle_get_tippy_top(C,request):
		A=request.path[C.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=web.Response();B.content_type=_E;B.text='{\n  "ietf-restconf:restconf" : {\n    "data" : {},\n    "operations" : {},\n    "yang-library-version" : "2019-01-04"\n  }\n}';return B
	async def handle_get_yang_library_version(C,request):
		D=request;A=D.path[C.len_prefix_running:]
		if A=='':A=_A
		elif A!=_A and A[-1]==_A:A=A[:-1]
		B=await C._check_auth(D,A)
		if B.status==401:return B
		B=await C.check_headers(D)
		if B!=_B:return B
		E=web.Response();E.content_type=_E;E.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}';return E


# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_G=':memory:'
_F='mysql'
_E='postgresql'
_D=True
_C='sqlite'
_B='postgres'
_A=None
import os
from copy import copy
import sqlalchemy as sa
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import OperationalError,ProgrammingError
def database_exists(url,connect_args=_A):
	I=False;E=connect_args;B=url
	def F(engine,sql):A=engine;B=A.execute(sql);C=B.scalar();B.close();A.dispose();return C
	def G(database):
		A=database
		if not os.path.isfile(A)or os.path.getsize(A)<100:return I
		with open(A,'rb')as B:C=B.read(100)
		return C[:16]==b'SQLite format 3\x00'
	B=copy(make_url(B));C=B.database
	if B.drivername.startswith(_B):B.database=_B
	else:B.database=_A
	A=sa.create_engine(B,connect_args=E)
	if A.dialect.name==_E:D="SELECT 1 FROM pg_database WHERE datname='%s'"%C;return bool(F(A,D))
	elif A.dialect.name==_F:D="SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '%s'"%C;return bool(F(A,D))
	elif A.dialect.name==_C:
		if C:return C==_G or G(C)
		else:return _D
	else:
		A.dispose();A=_A;D='SELECT 1'
		try:B.database=C;A=sa.create_engine(B,connect_args=E);H=A.execute(D);H.close();return _D
		except (ProgrammingError,OperationalError):return I
		finally:
			if A is not _A:A.dispose()
def create_database(url,encoding='utf8',template=_A,connect_args=_A):
	I=encoding;G=template;F=connect_args;B=url;B=copy(make_url(B));C=B.database
	if B.drivername.startswith(_B):B.database=_B
	elif B.drivername.startswith('mssql'):B.database='master'
	elif not B.drivername.startswith(_C):B.database=_A
	if B.drivername=='mssql+pyodbc':
		if F is not _A:H=copy(F)
		else:H={}
		H['autocommit']=_D;A=sa.create_engine(B,connect_args=H)
	elif B.drivername=='postgresql+pg8000':A=sa.create_engine(B,isolation_level='AUTOCOMMIT',connect_args=F)
	else:A=sa.create_engine(B,connect_args=F)
	D=_A
	if A.dialect.name==_E:
		if A.driver=='psycopg2':from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT as J;A.raw_connection().set_isolation_level(J)
		if not G:G='template1'
		E="CREATE DATABASE {0} ENCODING '{1}' TEMPLATE {2}".format(quote(A,C),I,quote(A,G));D=A.execute(E)
	elif A.dialect.name==_F:E="CREATE DATABASE {0} CHARACTER SET = '{1}'".format(quote(A,C),I);D=A.execute(E)
	elif A.dialect.name==_C and C!=_G:
		if C:A.execute('CREATE TABLE DB(id int);');A.execute('DROP TABLE DB;')
	else:E='CREATE DATABASE {0}'.format(quote(A,C));D=A.execute(E)
	if D is not _A:D.close()
	A.dispose()
from sqlalchemy.orm.session import object_session
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.orm.exc import UnmappedInstanceError
def quote(mixed,ident):
	A=mixed
	if isinstance(A,Dialect):B=A
	else:B=get_bind(A).dialect
	return B.preparer(B).quote(ident)
def get_bind(obj):
	A=obj
	if hasattr(A,'bind'):B=A.bind
	else:
		try:B=object_session(A).bind
		except UnmappedInstanceError:B=A
	if not hasattr(B,'execute'):raise TypeError('This method accepts only Session, Engine, Connection and declarative model objects.')
	return B